# -*- coding: utf-8 -*-
import numpy as np

from .parser import parse_pse_frame_to_alsep_words


class PseFrame(object):
    def __init__(self, frame, _is_old_format):
        self.data = np.require(frame, dtype=np.int64)
        self._is_old_format = _is_old_format
        # Frame header parameters
        self.software_time_flag = None
        self.msec_of_year = None
        self.alsep_tracking_station_id = None
        self.bit_error_rate = None
        self.data_rate = None
        self.alsep_word5 = None
        self.barker_code = None
        self.barker_code_comp = None
        self.frame_count = None
        self.mode_bit = None
        # Extracted ALSEP words
        self.alsep_words = None
        # Parse header and data
        self.parse_header()
        self.parse_data()

    def parse_header(self):
        self.software_time_flag = self.data[0] >> 7

        self.msec_of_year = np.int64(self.data[0] & 0x7f)
        self.msec_of_year = (self.msec_of_year << 8) + self.data[1]
        self.msec_of_year = (self.msec_of_year << 8) + self.data[2]
        self.msec_of_year = (self.msec_of_year << 8) + self.data[3]
        self.msec_of_year = (self.msec_of_year << 4) + (self.data[4] >> 4)

        self.alsep_tracking_station_id = self.data[4] & 0x0f
        self.bit_error_rate = (self.data[5] >> 2) & 0x3f
        self.data_rate = (self.data[5] >> 1) & 0x01

        self.alsep_word5 = self.data[6] & 0x03
        self.alsep_word5 = (self.alsep_word5 << 8) | self.data[7]

        self.barker_code = self.data[8]
        self.barker_code = (self.barker_code << 3) + (self.data[9] >> 5)

        self.barker_code_comp = self.data[9] & 0x0f
        self.barker_code_comp = \
            (self.barker_code_comp << 7) + (self.data[10] >> 1)

        self.frame_count = self.data[11] >> 1

        self.mode_bit = self.data[11] & 0x01

    def parse_data(self):
        self.alsep_words = \
            parse_pse_frame_to_alsep_words(self.data, self._is_old_format)
        self.alsep_words[5] = self.alsep_word5

    def dump(self):
        print('--- Frame info ---')
        print('software time flag: {}'.format(self.software_time_flag))
        print('msec of year: {:d}'.format(self.msec_of_year))
        print('ALSEP tracking station ID: {}'.format(
            self.alsep_tracking_station_id))
        print('bit Error Rate: {}'.format(self.bit_error_rate))
        print('data Rate: {}'.format(self.data_rate))
        print('ALSEP word 5: {}'.format(self.alsep_word5))
        print('sync pattern barker code: {}'.format(self.barker_code))
        print('sync pattern barker code compliment: {}'.format(
            self.barker_code_comp))
        print('mode bit: {}'.format(self.mode_bit))
        print('ALSEP words: {}'.format(self.alsep_words))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np

from obspy import UTCDateTime, read
from obspy.io.ah.core import _is_ah, _read_ah, _write_ah1, _read_ah1
from obspy.core.util import NamedTemporaryFile
import pytest


class TestCore():
    """
    AH (Ad Hoc) file test suite.
    """
    def test_is_ah(self, testdata, datapath):
        """
        Testing AH file format.
        """
        # AH v1
        testfile = datapath / 'TSG' / 'BRV.TSG.DS.lE21.resp'
        assert _is_ah(testfile)
        testfile = datapath / 'TSG' / 'BRV.TSG.KSM.sE12.resp'
        assert _is_ah(testfile)
        testfile = testdata['ah1.f']
        assert _is_ah(testfile)
        testfile = testdata['ah1.c']
        assert _is_ah(testfile)
        testfile = testdata['ah1.t']
        assert _is_ah(testfile)
        testfile = testdata['hrv.lh.zne']
        assert _is_ah(testfile)

        # AH v2
        testfile = testdata['ah2.f']
        assert _is_ah(testfile)
        testfile = testdata['ah2.f-e']
        assert _is_ah(testfile)
        testfile = testdata['ah2.c']
        assert _is_ah(testfile)
        testfile = testdata['ah2.t']
        assert _is_ah(testfile)

        # non AH files
        testfile = datapath / 'TSG' / 'BRV.TSG.DS.lE21.asc'
        assert not _is_ah(testfile)
        testfile = datapath / 'TSG' / 'BRV.TSG.KSM.sE12.asc'
        assert not _is_ah(testfile)
        testfile = datapath / 'TSG' / 'Readme_TSG_response.txt'
        assert not _is_ah(testfile)

    def test_read(self, testdata):
        """
        Testing reading AH file format using read() function.
        """
        # AH v1
        testfile = testdata['hrv.lh.zne']
        st = read(testfile)
        assert len(st) == 3
        testfile = testdata['ah1.f']
        st = read(testfile)
        assert len(st) == 4
        # not supported data types (vector, complex, tensor)
        testfile = testdata['ah1.c']
        with pytest.raises(NotImplementedError):
            _read_ah(testfile)
        testfile = testdata['ah1.t']
        with pytest.raises(NotImplementedError):
            _read_ah(testfile)

        # AH v2
        # float
        testfile = testdata['ah2.f']
        st = read(testfile)
        assert len(st) == 4

    def test_read_ah(self, testdata):
        """
        Testing reading AH file format using _read_ah() function.
        """
        # AH v1
        testfile = testdata['ah1.f']
        st = _read_ah(testfile)
        assert len(st) == 4
        tr = st[0]
        ah = tr.stats.ah
        # station
        assert ah.version == '1.0'
        assert ah.station.code == 'RSCP'
        assert ah.station.channel == 'IPZ'
        assert ah.station.type == 'null'
        assert round(abs(ah.station.latitude-35.599899), 6) == 0
        assert round(abs(ah.station.longitude--85.568802), 6) == 0
        assert ah.station.elevation == 481.0
        assert round(abs(ah.station.gain-64200.121094), 6) == 0
        assert len(ah.station.poles) == 24
        assert len(ah.station.zeros) == 7
        # event
        assert ah.event.latitude == 0.0
        assert ah.event.longitude == 0.0
        assert ah.event.depth == 0.0
        assert ah.event.origin_time is None
        assert ah.event.comment == 'null'
        # record
        assert ah.record.type == 1
        assert ah.record.ndata == 720
        assert tr.stats.npts == 720
        assert len(tr) == 720
        assert tr.data.dtype == np.float64
        assert round(abs(ah.record.delta-0.25), 6) == 0
        assert round(abs(tr.stats.delta-0.25), 6) == 0
        assert round(abs(ah.record.max_amplitude-0.0), 6) == 0
        dt = UTCDateTime(1984, 4, 20, 6, 42, 0, 120000)
        assert ah.record.start_time == dt
        assert tr.stats.starttime == dt
        assert ah.record.comment == 'null'
        assert ah.record.log == 'gdsn_tape;demeaned;'
        # data
        np.testing.assert_array_almost_equal(tr.data[:4], np.array([
            -731.41247559, -724.41247559, -622.41247559, -470.4125061]))
        np.testing.assert_array_almost_equal(tr.data[-4:], np.array([
            -1421.41247559, 118.58750153, 88.58750153, -982.41247559]))

        # not supported data types (vector, complex, tensor)
        testfile = testdata['ah1.c']
        with pytest.raises(NotImplementedError):
            _read_ah(testfile)
        testfile = testdata['ah1.t']
        with pytest.raises(NotImplementedError):
            _read_ah(testfile)

        # AH v2
        testfile = testdata['ah2.f']
        st = _read_ah(testfile)
        assert len(st) == 4
        tr = st[0]
        ah = tr.stats.ah
        assert ah.version == '2.0'
        # station
        assert ah.station.code == 'RSCP'
        assert ah.station.channel == 'IPZ'
        assert ah.station.type == 'null'
        assert round(abs(ah.station.latitude-35.599899), 6) == 0
        assert round(abs(ah.station.longitude--85.568802), 6) == 0
        assert ah.station.elevation == 481.0
        assert round(abs(ah.station.gain-64200.121094), 6) == 0
        assert len(ah.station.poles) == 24
        assert len(ah.station.zeros) == 7
        # event
        assert ah.event.latitude == 0.0
        assert ah.event.longitude == 0.0
        assert ah.event.depth == 0.0
        assert ah.event.origin_time is None
        assert ah.event.comment == 'null'
        # record
        assert ah.record.type == 1
        assert ah.record.ndata == 720
        assert tr.stats.npts == 720
        assert len(tr) == 720
        assert tr.data.dtype == np.float64
        assert round(abs(ah.record.delta-0.25), 6) == 0
        assert round(abs(tr.stats.delta-0.25), 6) == 0
        assert round(abs(ah.record.max_amplitude-0.0), 6) == 0
        dt = UTCDateTime(1984, 4, 20, 6, 42, 0, 120000)
        assert ah.record.start_time == dt
        assert tr.stats.starttime == dt
        assert ah.record.comment == 'null'
        assert ah.record.log == 'gdsn_tape;demeaned;'
        # data
        np.testing.assert_array_almost_equal(tr.data[:4], np.array([
            -731.41247559, -724.41247559, -622.41247559, -470.4125061]))
        np.testing.assert_array_almost_equal(tr.data[-4:], np.array([
            -1421.41247559, 118.58750153, 88.58750153, -982.41247559]))

        # not supported data types (vector, complex, tensor)
        testfile = testdata['ah2.t']
        with pytest.raises(NotImplementedError):
            _read_ah(testfile)

    def test_tsg(self, datapath):
        """
        Test reading AH v1 files of the STsR-TSG System at Borovoye.

        .. seealso:: https://www.ldeo.columbia.edu/res/pi/Monitoring/Data/
        """
        # 1 - BRV.TSG.DS.lE21
        testfile = datapath / 'TSG' / 'BRV.TSG.DS.lE21.resp'
        st = _read_ah(testfile)
        assert len(st) == 1
        tr = st[0]
        ah = tr.stats.ah
        assert ah.version == '1.0'
        # station
        assert ah.station.code == 'BRVK'
        assert ah.station.channel == 'lE21'
        assert ah.station.type == 'TSG-DS'
        assert round(abs(ah.station.latitude-53.058060), 6) == 0
        assert round(abs(ah.station.longitude-70.282799), 6) == 0
        assert ah.station.elevation == 300.0
        assert round(abs(ah.station.gain-0.05), 6) == 0
        assert round(abs(ah.station.normalization-40.009960), 6) == 0
        assert round(abs(ah.station.longitude-70.282799), 6) == 0
        # calibration
        assert len(ah.station.poles) == 7
        assert round(abs(
            ah.station.poles[0]-complex(-1.342653e-01, 1.168836e-01)), 6) == 0
        assert round(abs(
            ah.station.poles[1]-complex(-1.342653e-01, -1.168836e-01)), 6) == 0
        assert len(ah.station.zeros) == 4
        assert round(abs(ah.station.zeros[0]-complex(0.0, 0.0)), 6) == 0
        assert round(abs(ah.station.zeros[1]-complex(0.0, 0.0)), 6) == 0
        assert round(abs(ah.station.zeros[2]-complex(0.0, 0.0)), 6) == 0
        assert round(abs(ah.station.zeros[3]-complex(0.0, 0.0)), 6) == 0
        # event
        assert round(abs(ah.event.latitude-49.833000), 6) == 0
        assert round(abs(ah.event.longitude-78.807999), 6) == 0
        assert ah.event.depth == 0.5
        assert ah.event.origin_time == UTCDateTime(1988, 2, 8, 15, 23)
        assert ah.event.comment == 'Calibration_for_hg_TSG'
        # record
        assert ah.record.type == 1
        assert ah.record.ndata == 225
        assert tr.stats.npts == 225
        assert len(tr) == 225
        assert tr.data.dtype == np.float64
        assert round(abs(ah.record.delta-0.312), 6) == 0
        assert round(abs(tr.stats.delta-0.312), 6) == 0
        assert round(abs(ah.record.max_amplitude-785.805786), 6) == 0
        dt = UTCDateTime(1988, 2, 8, 15, 24, 50.136002)
        assert ah.record.start_time == dt
        assert tr.stats.starttime == dt
        assert ah.record.abscissa_min == 0.0
        assert ah.record.comment == 'DS response in counts/nm;'
        assert ah.record.log == \
            'brv2ah: ahtedit;demeaned;modhead;modhead;ahtedit;'
        # extras
        assert len(ah.extras) == 21
        assert ah.extras[0] == 0.0
        assert round(abs(ah.extras[1]-0.1), 6) == 0
        assert round(abs(ah.extras[2]-0.1), 6) == 0
        assert ah.extras[3] == 0.0
        # data
        np.testing.assert_array_almost_equal(tr.data[:24], np.array([
            -1.19425595, -1.19425595, -1.19425595, -1.19425595,
            -1.19425595, -1.19425595, -1.19425595, -1.19425595,
            -1.19425595, -1.19425595, -1.19425595, -1.19425595,
            -1.19425595, -1.19425595, -1.19425595, -1.19425595,
            -1.19425595, -1.19425595, -1.19425595, -1.19425595,
            52.8057518, 175.80580139, 322.80578613, 463.80578613]))
        np.testing.assert_array_almost_equal(tr.data[-4:], np.array([
            1.80574405, 2.80574393, 3.80574393, 3.80574393]))

    def test_write_ah1(self, testdata):
        """
        Testing writing AH1 file format using _write_ah1() function.
        """
        # AH v1
        testfile = testdata['st.ah']
        stream_orig = _read_ah(testfile)

        with NamedTemporaryFile() as tf:
            tmpfile = tf.name + '.AH'
            # write testfile
            _write_ah1(stream_orig, tmpfile)
            # read again
            st = _read_ah1(tmpfile)
            assert len(st) == 1
            tr = st[0]
            ah = tr.stats.ah
            stats = tr.stats
            # stream header
            assert stats.network == ''
            assert stats.station == 'ALE'
            assert stats.location == ''
            assert stats.channel == 'VHZ'
            starttime = UTCDateTime(1994, 6, 9, 0, 40, 45)
            endtime = UTCDateTime(1994, 6, 12, 8, 55, 4, 724522)
            assert stats.starttime == starttime
            assert stats.endtime == endtime
            assert round(abs(stats.sampling_rate-0.100000), 6) == 0
            assert round(abs(stats.delta-9.999990), 6) == 0
            assert stats.npts == 28887
            assert len(tr) == 28887
            assert stats.calib == 1.0

            # station
            assert ah.version == '1.0'
            assert ah.station.code == 'ALE'
            assert ah.station.channel == 'VHZ'
            assert ah.station.type == 'Global S'
            assert ah.station.latitude == 82.50330352783203
            assert ah.station.longitude == -62.349998474121094
            assert ah.station.elevation == 60.0
            assert ah.station.gain == 265302864.0
            assert len(ah.station.poles) == 13
            assert len(ah.station.zeros) == 6
            # event
            assert ah.event.latitude == -13.872200012207031
            assert ah.event.longitude == -67.51249694824219
            assert ah.event.depth == 640000.0
            origintime = UTCDateTime(1994, 6, 9, 0, 33, 16)
            assert ah.event.origin_time == origintime
            assert ah.event.comment == 'null'
            # record
            assert ah.record.type == 1
            assert ah.record.ndata == 28887
            assert tr.data.dtype == np.float64
            assert round(abs(ah.record.delta-9.999990), 6) == 0
            assert ah.record.max_amplitude == 9.265750885009766
            rstarttime = UTCDateTime(1994, 6, 9, 0, 40, 45)
            assert ah.record.start_time == rstarttime
            comment = 'Comp azm=0.0,inc=-90.0; Disp (m);'
            assert ah.record.comment == comment
            assert ah.record.log == 'null'
            # data
            np.testing.assert_array_almost_equal(tr.data[:4], np.array([
                -236., -242., -252., -262.]))
            np.testing.assert_array_almost_equal(tr.data[-4:], np.array([
                101., 106., 107., 104.]))



import numpy as np
import matplotlib.pyplot as plt
try:
    import mlpy
except ModuleNotFoundError:
    import warnings
    warnings.warn("mlpy not installed, code snippet skipped")
    exit(1)

import obspy
from obspy.imaging.cm import obspy_sequential


tr = obspy.read("https://examples.obspy.org/a02i.2008.240.mseed")[0]

omega0 = 8
wavelet_fct = "morlet"
scales = mlpy.wavelet.autoscales(N=len(tr.data), dt=tr.stats.delta, dj=0.05,
                                 wf=wavelet_fct, p=omega0)
spec = mlpy.wavelet.cwt(tr.data, dt=tr.stats.delta, scales=scales,
                        wf=wavelet_fct, p=omega0)
# approximate scales through frequencies
freq = (omega0 + np.sqrt(2.0 + omega0 ** 2)) / (4 * np.pi * scales[1:])

fig = plt.figure()
ax1 = fig.add_axes([0.1, 0.75, 0.7, 0.2])
ax2 = fig.add_axes([0.1, 0.1, 0.7, 0.60], sharex=ax1)
ax3 = fig.add_axes([0.83, 0.1, 0.03, 0.6])

t = np.arange(tr.stats.npts) / tr.stats.sampling_rate
ax1.plot(t, tr.data, 'k')

img = ax2.imshow(np.abs(spec), extent=[t[0], t[-1], freq[-1], freq[0]],
                 aspect='auto', interpolation='nearest', cmap=obspy_sequential)
# Hackish way to overlay a logarithmic scale over a linearly scaled image.
twin_ax = ax2.twinx()
twin_ax.set_yscale('log')
twin_ax.set_xlim(t[0], t[-1])
twin_ax.set_ylim(freq[-1], freq[0])
ax2.tick_params(which='both', labelleft=False, left=False)
twin_ax.tick_params(which='both', labelleft=True, left=True, labelright=False)

fig.colorbar(img, cax=ax3)

plt.show()



# -*- coding: utf-8 -*-
from .blockette import Blockette

# Volume Index Control Headers
from .blockette010 import Blockette010
from .blockette011 import Blockette011
from .blockette012 import Blockette012

# Abbreviation Dictionary Control Header
from .blockette030 import Blockette030
from .blockette031 import Blockette031
from .blockette032 import Blockette032
from .blockette033 import Blockette033
from .blockette034 import Blockette034
from .blockette041 import Blockette041
from .blockette043 import Blockette043
from .blockette044 import Blockette044
from .blockette047 import Blockette047
from .blockette048 import Blockette048

# Station Control Headers
from .blockette050 import Blockette050
from .blockette051 import Blockette051
from .blockette052 import Blockette052
from .blockette053 import Blockette053
from .blockette054 import Blockette054
from .blockette055 import Blockette055
from .blockette057 import Blockette057
from .blockette058 import Blockette058
from .blockette059 import Blockette059
from .blockette060 import Blockette060
from .blockette061 import Blockette061
from .blockette062 import Blockette062



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import pytest

from obspy.scripts.print import main as obspy_print
from obspy.core.util.misc import CatchOutput


class TestPrint():
    @pytest.fixture(scope='class')
    def all_files(self, root):
        all_files = [str(root / 'io' / 'ascii' / 'tests' / 'data' / name)
                     for name in ('slist.ascii', 'tspair.ascii')]
        return all_files

    def test_print(self, all_files):
        with CatchOutput() as out:
            obspy_print(all_files)

        expected = '''1 Trace(s) in Stream:
XX.TEST..BHZ | 2008-01-15T00:00:00.025000Z - 2008-01-15T00:00:15.875000Z | 40.0 Hz, 635 samples
'''  # noqa
        assert expected == out.stdout

    def test_print_nomerge(self, all_files):
        with CatchOutput() as out:
            obspy_print(['--no-merge'] + all_files)

        expected = '''2 Trace(s) in Stream:
XX.TEST..BHZ | 2008-01-15T00:00:00.025000Z - 2008-01-15T00:00:15.875000Z | 40.0 Hz, 635 samples
XX.TEST..BHZ | 2008-01-15T00:00:00.025000Z - 2008-01-15T00:00:15.875000Z | 40.0 Hz, 635 samples
'''  # noqa
        assert expected == out.stdout



# -*- coding: utf-8 -*-
"""
SAC bindings to ObsPy core module.

:copyright:
    The ObsPy Development Team (devs@obspy.org) & C. J. Ammon & J. MacCarthy
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import io
import os
import struct
from pathlib import Path

from obspy import Stream

from .sactrace import SACTrace


def _is_sac(filename):
    """
    Checks whether a file is a SAC file or not.

    :param filename: SAC file to be checked.
    :type filename: str, open file, or file-like object
    :rtype: bool
    :return: ``True`` if a SAC file.

    .. rubric:: Example

    >>> from obspy.core.util import get_example_file
    >>> _is_sac(get_example_file('test.sac'))
    True
    >>> _is_sac(get_example_file('test.mseed'))
    False
    """
    if isinstance(filename, io.BufferedIOBase):
        return _internal_is_sac(filename)
    elif isinstance(filename, (str, bytes)):
        with open(filename, "rb") as fh:
            return _internal_is_sac(fh)
    else:
        raise ValueError("Cannot open '%s'." % filename)


def _internal_is_sac(buf):
    """
    Checks whether a file-like object contains a SAC file or not.

    :param buf: SAC file to be checked.
    :type buf: file-like object or open file
    :rtype: bool
    :return: ``True`` if a SAC file.
    """
    starting_pos = buf.tell()
    try:
        # read delta (first header float)
        delta_bin = buf.read(4)
        delta = struct.unpack('<f', delta_bin)[0]
        # read nvhdr (70 header floats, 6 position in header integers)
        buf.seek(starting_pos + 4 * 70 + 4 * 6, 0)
        nvhdr_bin = buf.read(4)
        nvhdr = struct.unpack('<i', nvhdr_bin)[0]
        # read leven (70 header floats, 35 header integers, 0 position in
        # header bool)
        buf.seek(starting_pos + 4 * 70 + 4 * 35, 0)
        leven_bin = buf.read(4)
        leven = struct.unpack('<i', leven_bin)[0]
        # read lpspol (70 header floats, 35 header integers, 1 position in
        # header bool)
        buf.seek(starting_pos + 4 * 70 + 4 * 35 + 4 * 1, 0)
        lpspol_bin = buf.read(4)
        lpspol = struct.unpack('<i', lpspol_bin)[0]
        # read lovrok (70 header floats, 35 header integers, 2 position in
        # header bool)
        buf.seek(starting_pos + 4 * 70 + 4 * 35 + 4 * 2, 0)
        lovrok_bin = buf.read(4)
        lovrok = struct.unpack('<i', lovrok_bin)[0]
        # read lcalda (70 header floats, 35 header integers, 3 position in
        # header bool)
        buf.seek(starting_pos + 4 * 70 + 4 * 35 + 4 * 3, 0)
        lcalda_bin = buf.read(4)
        lcalda = struct.unpack('<i', lcalda_bin)[0]
        # check if file is big-endian
        if nvhdr < 0 or nvhdr > 20:
            nvhdr = struct.unpack('>i', nvhdr_bin)[0]
            delta = struct.unpack('>f', delta_bin)[0]
            leven = struct.unpack('>i', leven_bin)[0]
            lpspol = struct.unpack('>i', lpspol_bin)[0]
            lovrok = struct.unpack('>i', lovrok_bin)[0]
            lcalda = struct.unpack('>i', lcalda_bin)[0]
        # check again nvhdr
        if nvhdr < 1 or nvhdr > 20:
            return False
        if delta <= 0:
            return False
        if leven != 0 and leven != 1 and leven != -12345:
            return False
        if lpspol != 0 and lpspol != 1 and lpspol != -12345:
            return False
        if lovrok != 0 and lovrok != 1 and lovrok != -12345:
            return False
        if lcalda != 0 and lcalda != 1 and lcalda != -12345:
            return False
    except Exception:
        return False
    finally:
        # Reset buffer head position after reading.
        buf.seek(starting_pos, 0)
    return True


def _is_sac_xy(filename):
    """
    Checks whether a file is alphanumeric SAC file or not.

    :param filename: Alphanumeric SAC file to be checked.
    :type filename: str, open file, or file-like object
    :rtype: bool
    :return: ``True`` if a alphanumeric SAC file.

    .. rubric:: Example

    >>> from obspy.core.util import get_example_file
    >>> _is_sac_xy(get_example_file('testxy.sac'))
    True
    >>> _is_sac_xy(get_example_file('test.sac'))
    False
    """
    if isinstance(filename, io.BufferedIOBase):
        return _internal_is_sac_xy(filename)
    elif isinstance(filename, (str, bytes)):
        with open(filename, "rb") as fh:
            return _internal_is_sac_xy(fh)
    else:
        raise ValueError("Cannot open '%s'." % filename)


def _internal_is_sac_xy(buf):
    """
    Checks whether a file is alphanumeric SAC file or not.

    :param buf: Alphanumeric SAC file to be checked.
    :type buf: file-like object or open file
    :rtype: bool
    :return: ``True`` if a alphanumeric SAC file.
    """
    cur_pos = buf.tell()
    try:
        try:
            hdcards = []
            # read in the header cards
            for _i in range(30):
                hdcards.append(buf.readline())
            npts = int(hdcards[15].split()[-1])
            # read in the seismogram
            seis = buf.read(-1).split()
        except Exception:
            return False
        # check that npts header value and seismogram length are consistent
        if npts != len(seis):
            return False
        return True
    finally:
        buf.seek(cur_pos, 0)


def _read_sac_xy(filename, headonly=False, debug_headers=False,
                 round_sampling_interval=True, **kwargs):  # @UnusedVariable
    """
    Reads an alphanumeric SAC file and returns an ObsPy Stream object.

    .. warning::
        This function should NOT be called directly, it registers via the
        ObsPy :func:`~obspy.core.stream.read` function, call this instead.

    :param filename: Alphanumeric SAC file to be read.
    :type filename: str, :class:`~pathlib.Path`, open file, or file-like object
    :param headonly: If set to True, read only the head. This is most useful
        for scanning available data in huge (temporary) data sets.
    :type headonly: bool
    :param debug_headers: Extracts also the SAC headers ``'nzyear', 'nzjday',
        'nzhour', 'nzmin', 'nzsec', 'nzmsec', 'delta', 'scale', 'npts',
        'knetwk', 'kstnm', 'kcmpnm'`` which are usually directly mapped to the
        :class:`~obspy.core.stream.Stream` object if set to ``True``. Those
        values are not synchronized with the Stream object itself and won't
        be used during writing of a SAC file! Defaults to ``False``.
    :type debug_headers: bool
    :param round_sampling_interval: Whether to round sampling interval to
        microseconds before calculating sampling rate to avoid floating point
        accuracy issues with some SAC files (see #3408)
    :type round_sampling_interval: bool
    :rtype: :class:`~obspy.core.stream.Stream`
    :return: A ObsPy Stream object.

    .. rubric:: Example

    >>> from obspy import read
    >>> st = read("/path/to/testxy.sac")
    """
    if isinstance(filename, io.BufferedIOBase):
        return _internal_read_sac_xy(
            buf=filename, headonly=headonly, debug_headers=debug_headers,
            round_sampling_interval=round_sampling_interval, **kwargs)
    else:
        with open(filename, "rb") as fh:
            return _internal_read_sac_xy(
                buf=fh, headonly=headonly, debug_headers=debug_headers,
                round_sampling_interval=round_sampling_interval, **kwargs)


def _internal_read_sac_xy(
        buf, headonly=False, debug_headers=False, round_sampling_interval=True,
        **kwargs):  # @UnusedVariable
    """
    Reads an alphanumeric SAC file and returns an ObsPy Stream object.

    .. warning::
        This function should NOT be called directly, it registers via the
        ObsPy :func:`~obspy.core.stream.read` function, call this instead.

    :param buf: Alphanumeric SAC file to be read.
    :type buf: file or file-like object
    :param headonly: If set to True, read only the head. This is most useful
        for scanning available data in huge (temporary) data sets.
    :type headonly: bool
    :param debug_headers: Extracts also the SAC headers ``'nzyear', 'nzjday',
        'nzhour', 'nzmin', 'nzsec', 'nzmsec', 'delta', 'scale', 'npts',
        'knetwk', 'kstnm', 'kcmpnm'`` which are usually directly mapped to the
        :class:`~obspy.core.stream.Stream` object if set to ``True``. Those
        values are not synchronized with the Stream object itself and won't
        be used during writing of a SAC file! Defaults to ``False``.
    :type debug_headers: bool
    :param round_sampling_interval: Whether to round sampling interval to
        microseconds before calculating sampling rate to avoid floating point
        accuracy issues with some SAC files (see #3408)
    :type round_sampling_interval: bool
    :rtype: :class:`~obspy.core.stream.Stream`
    :return: A ObsPy Stream object.

    .. rubric:: Example

    >>> from obspy import read
    >>> st = read("/path/to/testxy.sac")
    """
    sac = SACTrace.read(buf, headonly=headonly, ascii=True)
    # assign all header entries to a new dictionary compatible with ObsPy
    tr = sac.to_obspy_trace(debug_headers=debug_headers,
                            round_sampling_interval=round_sampling_interval)

    return Stream([tr])


def _write_sac_xy(stream, filename, **kwargs):  # @UnusedVariable
    """
    Writes a alphanumeric SAC file.

    .. warning::
        This function should NOT be called directly, it registers via the
        the :meth:`~obspy.core.stream.Stream.write` method of an
        ObsPy :class:`~obspy.core.stream.Stream` object, call this instead.

    :param stream: The ObsPy Stream object to write.
    :type stream: :class:`~obspy.core.stream.Stream`
    :param filename: Name of file to write. In case an open file or
        file-like object is passed, this function only supports writing
        Stream objects containing a single Trace. This is a limitation of
        the SAC file format. An exception will be raised in case it's
        necessary.
    :type filename: str, open file, or file-like object

    .. rubric:: Example

    >>> from obspy import read
    >>> st = read()
    >>> st.write("testxy.sac", format="SACXY")  #doctest: +SKIP
    """
    # SAC can only store one Trace per file.
    if isinstance(filename, io.BufferedIOBase):
        if len(stream) > 1:
            raise ValueError("If writing to a file-like object in the SAC "
                             "format, the Stream object can only contain "
                             "one Trace.")
        _internal_write_sac_xy(stream[0], filename, **kwargs)
        return
    elif isinstance(filename, (str, bytes)):
        # Otherwise treat it as a filename
        # Translate the common (renamed) entries
        base, ext = os.path.splitext(filename)
        for i, trace in enumerate(stream):
            if len(stream) != 1:
                filename = "%s%02d%s" % (base, i + 1, ext)
            with open(filename, "wb") as fh:
                _internal_write_sac_xy(trace, fh, **kwargs)
    else:
        raise ValueError("Cannot open '%s'." % filename)


def _internal_write_sac_xy(trace, buf, **kwargs):  # @UnusedVariable
    """
    Writes a single trace to alphanumeric SAC file.

    .. warning::
        This function should NOT be called directly, it registers via the
        the :meth:`~obspy.core.stream.Stream.write` method of an
        ObsPy :class:`~obspy.core.stream.Stream` object, call this instead.

    :param trace: The ObsPy Trace object to write.
    :type trace: :class:`~obspy.core.trace.Trace`
    :param buf: Object to write to.
    :type buf: file-like object
    """
    sac = SACTrace.from_obspy_trace(trace, keep_sac_header=True)
    sac.write(buf, ascii=True, flush_headers=False)


def _read_sac(filename, headonly=False, debug_headers=False, fsize=True,
              round_sampling_interval=True, **kwargs):  # @UnusedVariable
    """
    Reads an SAC file and returns an ObsPy Stream object.

    .. warning::
        This function should NOT be called directly, it registers via the
        ObsPy :func:`~obspy.core.stream.read` function, call this instead.

    :param filename: SAC file to be read.
    :type filename: str, :class:`~pathlib.Path`, open file, or file-like object
    :param headonly: If set to True, read only the head. This is most useful
        for scanning available data in huge (temporary) data sets.
    :type headonly: bool
    :param debug_headers: Extracts also the SAC headers ``'nzyear', 'nzjday',
        'nzhour', 'nzmin', 'nzsec', 'nzmsec', 'delta', 'scale', 'npts',
        'knetwk', 'kstnm', 'kcmpnm'`` which are usually directly mapped to the
        :class:`~obspy.core.stream.Stream` object if set to ``True``. Those
        values are not synchronized with the Stream object itself and won't
        be used during writing of a SAC file! Defaults to ``False``.
    :type debug_headers: bool
    :param fsize: Check if file size is consistent with theoretical size
        from header. Defaults to ``True``.
    :type fsize: bool
    :param round_sampling_interval: Whether to round sampling interval to
        microseconds before calculating sampling rate to avoid floating point
        accuracy issues with some SAC files (see #3408)
    :type round_sampling_interval: bool
    :rtype: :class:`~obspy.core.stream.Stream`
    :return: A ObsPy Stream object.

    .. rubric:: Example

    >>> from obspy import read
    >>> st = read("/path/to/test.sac")
    """
    # Only byte buffers for binary SAC.
    if isinstance(filename, io.BufferedIOBase):
        return _internal_read_sac(
            buf=filename, headonly=headonly, debug_headers=debug_headers,
            fsize=fsize, round_sampling_interval=round_sampling_interval,
            **kwargs)
    elif isinstance(filename, (str, bytes, Path)):
        with open(filename, "rb") as fh:
            return _internal_read_sac(
                buf=fh, headonly=headonly, debug_headers=debug_headers,
                fsize=fsize, round_sampling_interval=round_sampling_interval,
                **kwargs)
    else:
        raise ValueError("Cannot open '%s'." % filename)


def _internal_read_sac(buf, headonly=False, debug_headers=False, fsize=True,
                       byteorder=None, round_sampling_interval=True,
                       **kwargs):  # @UnusedVariable
    """
    Reads an SAC file and returns an ObsPy Stream object.

    .. warning::
        This function should NOT be called directly, it registers via the
        ObsPy :func:`~obspy.core.stream.read` function, call this instead.

    :param buf: SAC file to be read.
    :type buf: file or file-like object
    :param headonly: If set to True, read only the head. This is most useful
        for scanning available data in huge (temporary) data sets.
    :type headonly: bool
    :param debug_headers: Extracts also the SAC headers ``'nzyear', 'nzjday',
        'nzhour', 'nzmin', 'nzsec', 'nzmsec', 'delta', 'scale', 'npts',
        'knetwk', 'kstnm', 'kcmpnm'`` which are usually directly mapped to the
        :class:`~obspy.core.stream.Stream` object if set to ``True``. Those
        values are not synchronized with the Stream object itself and won't
        be used during writing of a SAC file! Defaults to ``False``.
    :type debug_headers: bool
    :param fsize: Check if file size is consistent with theoretical size
        from header. Defaults to ``True``.
    :type fsize: bool
    :param byteorder: If omitted or None, automatic byte-order checking is
        done, starting with native order. If byteorder is specified,
        {'little', 'big'}, and incorrect, a
        :class:`obspy.io.sac.util.SacIOError` is raised. Only valid for binary
        files.
    :type byteorder: str, optional
    :param round_sampling_interval: Whether to round sampling interval to
        microseconds before calculating sampling rate to avoid floating point
        accuracy issues with some SAC files (see #3408)
    :type round_sampling_interval: bool
    :rtype: :class:`~obspy.core.stream.Stream`
    :return: A ObsPy Stream object.
    """

    # Extract encoding flag (default to ascii)
    encoding_str = kwargs.get('encoding', 'ascii')

    # read SAC file
    sac = SACTrace.read(buf, headonly=headonly, ascii=False,
                        byteorder=byteorder, checksize=fsize,
                        encoding=encoding_str)
    # assign all header entries to a new dictionary compatible with an ObsPy
    tr = sac.to_obspy_trace(debug_headers=debug_headers, encoding=encoding_str,
                            round_sampling_interval=round_sampling_interval)

    return Stream([tr])


def _write_sac(stream, filename, byteorder="<", **kwargs):  # @UnusedVariable
    """
    Writes a SAC file.

    .. warning::
        This function should NOT be called directly, it registers via the
        the :meth:`~obspy.core.stream.Stream.write` method of an
        ObsPy :class:`~obspy.core.stream.Stream` object, call this instead.

    :param stream: The ObsPy Stream object to write.
    :type stream: :class:`~obspy.core.stream.Stream`
    :param filename: Name of file to write. In case an open file or
        file-like object is passed, this function only supports writing
        Stream objects containing a single Trace. This is a limitation of
        the SAC file format. An exception will be raised in case it's
        necessary.
    :type filename: str, open file, or file-like object
    :param byteorder: Must be either ``0`` or ``'<'`` for LSBF or
        little-endian, ``1`` or ``'>'`` for MSBF or big-endian.
        Defaults to little endian.
    :type byteorder: int or str

    .. rubric:: Example

    >>> from obspy import read
    >>> st = read()
    >>> st.write("test.sac", format="SAC")  #doctest: +SKIP
    """
    # Bytes buffer are ok, but only if the Stream object contains only one
    # Trace. SAC can only store one Trace per file.
    if isinstance(filename, io.BufferedIOBase):
        if len(stream) > 1:
            raise ValueError("If writing to a file-like object in the SAC "
                             "format, the Stream object can only contain "
                             "one Trace.")
        _internal_write_sac(stream[0], filename, byteorder=byteorder, **kwargs)
        return
    elif isinstance(filename, (str, bytes)):
        # Otherwise treat it as a filename
        # Translate the common (renamed) entries
        base, ext = os.path.splitext(filename)
        for i, trace in enumerate(stream):
            if len(stream) != 1:
                filename = "%s%02d%s" % (base, i + 1, ext)
            with open(filename, "wb") as fh:
                _internal_write_sac(trace, fh, byteorder=byteorder, **kwargs)
    else:
        raise ValueError("Cannot open '%s'." % filename)


def _internal_write_sac(trace, buf, byteorder="<", keep_sac_header=True,
                        **kwargs):
    """
    Writes a single trace to an open file or file-like object

    .. warning::
        This function should NOT be called directly, it registers via the
        the :meth:`~obspy.core.stream.Stream.write` method of an
        ObsPy :class:`~obspy.core.stream.Stream` object, call this instead.

    :param trace: The ObsPy Trace object to write.
    :type trace: :class:`~obspy.core.trace.Trace`
    :param buf: Object to write to.
    :type buf: open file or file-like object
    :param byteorder: Must be either ``0`` or ``'<'`` for LSBF or
        little-endian, ``1`` or ``'>'`` for MSBF or big-endian.
        Defaults to little endian.
    :type byteorder: int or str
    :param keep_sac_header: Whether to merge the ``Stats`` header with an
        existing ``Stats.sac`` SAC header, if one exists. Defaults to ``True``.
        See :func:`~obspy.io.sac.util.obspy_to_sac_header` for more.
    :type keep_sac_header: bool
    """
    if byteorder in ("<", 0, "0"):
        byteorder = 'little'
    elif byteorder in (">", 1, "1"):
        byteorder = 'big'
    else:
        msg = "Invalid byte order. It must be either '<', '>', 0 or 1"
        raise ValueError(msg)

    sac = SACTrace.from_obspy_trace(trace, keep_sac_header)
    sac.write(buf, ascii=False, byteorder=byteorder)



# -*- coding: utf-8 -*-
"""
The obspy.imaging.scripts.plot / obspy-plot test suite.
"""
import shutil
from pathlib import Path

import pytest

from obspy.core.util.misc import TemporaryWorkingDirectory
from obspy.imaging.scripts.plot import main as obspy_plot


class TestPlot:
    """
    Test cases for obspy-plot
    """
    @pytest.fixture(scope='class')
    def all_files(self, root):
        """Collect all files. """
        all_files = [root / 'io' / 'ascii' / 'tests' / 'data' / i
                     for i in ['slist.ascii', 'slist_2_traces.ascii']]
        return all_files

    def test_plot(self, image_path, all_files):
        """
        Run obspy-plot on selected tests
        """
        # Copy files to a temp folder to avoid wildcard scans.
        with TemporaryWorkingDirectory():
            all_files_list = []
            for filename in all_files:
                newname = Path('.') / filename.name
                shutil.copy(filename, newname)
                all_files_list.append(str(newname))

            obspy_plot(['--outfile', str(image_path)] + all_files_list)

    def test_plot_no_merge(self, image_path, all_files):
        """
        Run obspy-plot without trace merging
        """
        # Copy files to a temp folder to avoid wildcard scans.
        with TemporaryWorkingDirectory():
            all_files_list = []
            for filename in all_files:
                newname = Path('.') / filename.name
                shutil.copy(filename, newname)
                all_files_list.append(str(newname))

            obspy_plot(['--no-automerge', '--outfile', str(image_path)] +
                       all_files_list)



# -*- coding: utf-8 -*-
import copy
import io
import pickle
import pytest
import warnings

from obspy import Stream, Trace, UTCDateTime, read
from obspy.core import Stats
from obspy.core.util.base import CatchAndAssertWarnings
from obspy.core.util import AttribDict


class TestStats:
    """
    Test suite for obspy.core.util.Stats.
    """
    nslc = ['network', 'station', 'location', 'channel']

    def test_init(self):
        """
        Init tests.
        """
        stats = Stats({'test': 'muh'})
        stats['other1'] = {'test1': '1'}
        stats['other2'] = AttribDict({'test2': '2'})
        stats['other3'] = 'test3'
        assert stats.test == 'muh'
        assert stats['test'] == 'muh'
        assert stats.other1.test1 == '1'
        assert stats.other1.__class__ == AttribDict
        assert len(stats.other1) == 1
        assert stats.other2.test2 == '2'
        assert stats.other2.__class__ == AttribDict
        assert len(stats.other2) == 1
        assert stats.other3 == 'test3'
        assert 'test' in stats
        assert 'test' in stats.__dict__

    def test_deepcopy(self):
        """
        Tests deepcopy method of Stats object.
        """
        stats = Stats()
        stats.network = 'BW'
        stats['station'] = 'ROTZ'
        stats['other1'] = {'test1': '1'}
        stats['other2'] = AttribDict({'test2': '2'})
        stats['other3'] = 'test3'
        stats2 = copy.deepcopy(stats)
        stats.network = 'CZ'
        stats.station = 'RJOB'
        assert stats2.__class__ == Stats
        assert stats2.network == 'BW'
        assert stats2.station == 'ROTZ'
        assert stats2.other1.test1 == '1'
        assert stats2.other1.__class__ == AttribDict
        assert len(stats2.other1) == 1
        assert stats2.other2.test2 == '2'
        assert stats2.other2.__class__ == AttribDict
        assert len(stats2.other2) == 1
        assert stats2.other3 == 'test3'
        assert stats.network == 'CZ'
        assert stats.station == 'RJOB'

    def test_update(self):
        """
        Tests update method of Stats object.
        """
        x = Stats({'a': 5})
        assert 'a' in dir(x)
        x.update({'b': 5})
        assert 'b' in dir(x)
        y = {'a': 5}
        y.update({'b': 5})
        x = Stats(y)
        assert 'b' in dir(x)

    def test_simple_stats(self):
        """
        Various setter and getter tests.
        """
        stats = Stats()
        stats.test = 1
        assert stats.test == 1
        assert stats['test'] == 1
        stats['test2'] = 2
        assert stats.test2 == 2
        assert stats['test2'] == 2
        stats['test'] = 2
        assert stats.test == 2
        assert stats['test'] == 2
        stats.test2 = 1
        assert stats.test2 == 1
        assert stats['test2'] == 1

    def test_nested_stats(self):
        """
        Various setter and getter tests.
        """
        # 1
        stats = Stats()
        stats.test = dict()
        stats.test['test2'] = 'muh'
        assert stats.test.test2 == 'muh'
        assert stats.test['test2'] == 'muh'
        assert stats['test'].test2 == 'muh'
        assert stats['test']['test2'] == 'muh'
        stats.test['test2'] = 'maeh'
        assert stats.test.test2 == 'maeh'
        assert stats.test['test2'] == 'maeh'
        assert stats['test'].test2 == 'maeh'
        assert stats['test']['test2'] == 'maeh'
        # 2 - multiple initialization
        stats = Stats({'muh': 'meah'})
        stats2 = Stats(Stats(Stats(stats)))
        assert stats2.muh == 'meah'
        # 3 - check conversion to AttribDict
        stats = Stats()
        stats.sub1 = {'muh': 'meah'}
        stats.sub2 = AttribDict({'muh2': 'meah2'})
        stats2 = Stats(stats)
        assert isinstance(stats.sub1, AttribDict)
        assert isinstance(stats.sub2, AttribDict)
        assert stats2.sub1.muh == 'meah'
        assert stats2.sub2.muh2 == 'meah2'

    def test_bugfix_set_stats(self):
        """
        Test related to issue #4.
        """
        st = Stream([Trace()])
        st += st
        # change stats attributes
        st[0].stats.station = 'AAA'
        st[1].stats['station'] = 'BBB'
        assert st[0].stats.station == 'BBB'
        assert st[0].stats['station'] == 'BBB'
        assert st[1].stats['station'] == 'BBB'
        assert st[1].stats.station == 'BBB'

    def test_bugfix_set_stats_2(self):
        """
        Second test related to issue #4.
        """
        st = Stream([Trace(header={'station': 'BGLD'})])
        assert st[0].stats.station == 'BGLD'
        assert st[0].stats['station'] == 'BGLD'
        st[0].stats.station = 'AAA'
        assert st[0].stats.station == 'AAA'
        assert st[0].stats['station'] == 'AAA'
        st = st + st
        assert st[0].stats.station == 'AAA'
        assert st[0].stats['station'] == 'AAA'
        st[0].stats.station = 'BBB'
        assert st[0].stats.station == 'BBB'
        assert st[0].stats['station'] == 'BBB'

    def test_bugfix_set_stats_3(self):
        """
        Third test related to issue #4.
        """
        st = Stream([Trace(header={'station': 'BGLD'})])
        assert st[0].stats.station == 'BGLD'
        st = st + st
        st[0].stats.station = 'AAA'
        st = st + st
        st[3].stats.station = 'BBB'
        # changed in rev. 1625: adding streams doesn't deepcopy
        # therefore all traces in the test stream are identical
        # (python list behavior)
        for tr in st:
            assert tr == st[0]
            assert tr.stats.station == 'BBB'
            assert tr.stats['station'] == 'BBB'
            assert tr.stats.get('station') == 'BBB'
            assert 'BBB' in tr.stats.values()

    def test_pickle_stats(self):
        """
        Test pickling Stats objects. Test case for issue #10.
        """
        stats = Stats()
        stats.muh = 1
        stats['maeh'] = 'hallo'
        # ASCII
        temp = pickle.dumps(stats, protocol=0)
        stats2 = pickle.loads(temp)
        assert stats == stats2
        # old binary
        temp = pickle.dumps(stats, protocol=1)
        stats2 = pickle.loads(temp)
        assert stats == stats2
        # new binary
        temp = pickle.dumps(stats, protocol=2)
        stats2 = pickle.loads(temp)
        assert stats == stats2
        # SOH channels sampling_rate & delta == 0. for #1989
        stats.sampling_rate = 0
        pickle.loads(pickle.dumps(stats, protocol=0))
        pickle.loads(pickle.dumps(stats, protocol=1))
        pickle.loads(pickle.dumps(stats, protocol=2))

    def test_set_calib(self):
        """
        Test to prevent setting a calibration factor of 0
        """
        x = Stats()
        # this should work
        x.update({'calib': 1.23})
        assert x.calib, 1.23
        # this raises UserWarning
        with warnings.catch_warnings(record=True):
            warnings.simplefilter('error', UserWarning)
            # 1
            with pytest.raises(UserWarning):
                x.__setitem__('calib', 0)
            # 2
            with pytest.raises(UserWarning):
                x.update({'calib': 0})
        # calib value should nevertheless be set to 0
        assert x.calib, 0

    def test_compare_with_dict(self):
        """
        Checks if Stats is still comparable to a dict object.
        """
        adict = {
            'network': '', 'sampling_rate': 1.0, 'test': 1, 'station': '',
            'location': '', 'starttime': UTCDateTime(1970, 1, 1, 0, 0),
            'delta': 1.0, 'calib': 1.0, 'npts': 0,
            'endtime': UTCDateTime(1970, 1, 1, 0, 0), 'channel': ''}
        ad = Stats(adict)
        assert ad == adict
        assert adict == ad

    def test_delta_zero(self):
        """
        Make sure you can set delta = 0. for #1989
        """
        stat = Stats()
        stat.delta = 0

    def test_non_str_in_nscl_raise_warning(self):
        """
        Ensure assigning a non-str value to network, station, location, or
        channel issues a warning, then casts value into str. See issue # 1995
        """
        stats = Stats()

        for val in self.nslc:
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter('default')
                setattr(stats, val, 42)
            # make sure a warning was issued
            assert len(w) == 1
            exp_str = 'Attribute "%s" must be of type ' % val
            assert exp_str in str(w[-1].message)
            # make sure the value was cast to a str
            new_val = getattr(stats, val)
            assert new_val == '42'

    def test_nscl_cannot_be_none(self):
        """
        Ensure the nslc values can't be assigned to None but rather None
        gets converted to a str
        """
        stats = Stats()
        for val in self.nslc:
            with pytest.warns(UserWarning):
                setattr(stats, val, None)
            assert getattr(stats, val) == 'None'

    @pytest.mark.filterwarnings('ignore:Attribute')
    def test_casted_stats_nscl_writes_to_mseed(self):
        """
        Ensure a Stream object that has had its nslc types cast to str can
        still be written.
        """
        st = Stream(traces=read()[0])

        # Get a new stats object with just the basic items in it
        stats_items = set(Stats())
        new_stats = Stats()
        new_stats.__dict__.update({x: st[0].stats[x] for x in stats_items})
        with CatchAndAssertWarnings():
            new_stats.network = 1
            new_stats.station = 1.1
        new_stats.channel = 'Non'
        st[0].stats = new_stats
        # try writing stream to bytes buffer
        bio = io.BytesIO()
        st.write(bio, 'mseed')
        bio.seek(0)
        # read bytes and compare
        stt = read(bio)
        # remove _mseed so streams can compare equal
        stt[0].stats.pop('mseed')
        del stt[0].stats._format  # format gets added upon writing
        assert st == stt

    @pytest.mark.filterwarnings('ignore:Attribute')
    def test_different_string_types(self):
        """
        Test the various types of strings found in the wild get converted to
        native_str type.
        """
        nbytes = bytes('HHZ', 'utf8')
        the_strs = ['HHZ', nbytes, u'HHZ']

        stats = Stats()

        for a_str in the_strs:
            for nslc in self.nslc:
                setattr(stats, nslc, a_str)
                assert isinstance(getattr(stats, nslc), str)

    def test_component(self):
        """
        Test setting and getting of component.
        """
        stats = Stats()
        # Channel with 3 characters
        stats.channel = 'HHZ'
        assert stats.component == 'Z'
        stats.component = 'L'
        assert stats.component == 'L'
        assert stats.channel == 'HHL'
        stats['component'] = 'Q'
        assert stats['component'] == 'Q'
        assert stats.channel == 'HHQ'
        # Channel with 1 character as component
        stats.channel = 'N'
        stats.component = 'E'
        assert stats.channel == 'E'
        assert stats.component == 'E'
        # Channel with 0 characters
        stats.channel = ''
        assert stats.component == ''
        stats.component = 'Z'
        assert stats.channel == 'Z'
        # Components must be single character
        stats.channel = 'HHZ'
        with pytest.raises(ValueError):
            stats.component = ''
        assert stats.channel == 'HHZ'
        with pytest.raises(ValueError):
            stats.component = 'ZZ'
        assert stats.channel == 'HHZ'



# -*- coding: utf-8 -*-
"""
The obspy.realtime.rttrace test suite.
"""
import warnings

import numpy as np

from obspy import Trace
from obspy.core.stream import read
from obspy.realtime import RtTrace
from obspy.realtime.rtmemory import RtMemory
import obspy.signal.filter
import pytest


class TestRtTrace():

    def test_eq(self):
        """
        Testing __eq__ method.
        """
        tr = Trace()
        tr2 = RtTrace()
        tr3 = RtTrace()
        # RtTrace should never be equal with Trace objects
        assert not (tr2 == tr)
        assert not tr2.__eq__(tr)
        assert tr2 == tr3
        assert tr2.__eq__(tr3)

    def test_ne(self):
        """
        Testing __ne__ method.
        """
        tr = Trace()
        tr2 = RtTrace()
        tr3 = RtTrace()
        # RtTrace should never be equal with Trace objects
        assert tr2 != tr
        assert tr2.__ne__(tr)
        assert not (tr2 != tr3)
        assert not tr2.__ne__(tr3)

    def test_register_rt_process(self):
        """
        Testing register_rt_process method.
        """
        tr = RtTrace()
        # 1 - function call
        tr.register_rt_process(np.abs)
        assert tr.processing == [(np.abs, {}, None)]
        # 2 - predefined RT processing algorithm
        tr.register_rt_process('integrate', test=1, muh='maeh')
        assert tr.processing[1][0] == 'integrate'
        assert tr.processing[1][1] == {'test': 1, 'muh': 'maeh'}
        assert isinstance(tr.processing[1][2][0], RtMemory)
        # 3 - contained name of predefined RT processing algorithm
        tr.register_rt_process('in')
        assert tr.processing[2][0] == 'integrate'
        tr.register_rt_process('integ')
        assert tr.processing[3][0] == 'integrate'
        tr.register_rt_process('integr')
        assert tr.processing[4][0] == 'integrate'
        # 4 - unknown functions
        with pytest.raises(NotImplementedError):
            tr.register_rt_process('integrate2')
        with pytest.raises(NotImplementedError):
            tr.register_rt_process('xyz')
        # 5 - module instead of function
        with pytest.raises(NotImplementedError):
            tr.register_rt_process(np)
        # check number off all processing steps within RtTrace
        assert len(tr.processing) == 5
        # check tr.stats.processing
        assert len(tr.stats.processing) == 5
        assert tr.stats.processing[0].startswith("realtime_process")
        assert 'absolute' in tr.stats.processing[0]
        for i in range(1, 5):
            assert 'integrate' in tr.stats.processing[i]
        # check kwargs
        assert "maeh" in tr.stats.processing[1]

    def test_append_sanity_checks(self):
        """
        Testing sanity checks of append method.
        """
        rtr = RtTrace()
        ftr = Trace(data=np.array([0, 1]))
        # sanity checks need something already appended
        rtr.append(ftr)
        # 1 - differing ID
        tr = Trace(header={'network': 'xyz'})
        with pytest.raises(TypeError):
            rtr.append(tr)
        tr = Trace(header={'station': 'xyz'})
        with pytest.raises(TypeError):
            rtr.append(tr)
        tr = Trace(header={'location': 'xy'})
        with pytest.raises(TypeError):
            rtr.append(tr)
        tr = Trace(header={'channel': 'xyz'})
        with pytest.raises(TypeError):
            rtr.append(tr)
        # 2 - sample rate
        tr = Trace(header={'sampling_rate': 100.0})
        with pytest.raises(TypeError):
            rtr.append(tr)
        tr = Trace(header={'delta': 0.25})
        with pytest.raises(TypeError):
            rtr.append(tr)
        # 3 - calibration factor
        tr = Trace(header={'calib': 100.0})
        with pytest.raises(TypeError):
            rtr.append(tr)
        # 4 - data type
        tr = Trace(data=np.array([0.0, 1.1]))
        with pytest.raises(TypeError):
            rtr.append(tr)
        # 5 - only Trace objects are allowed
        with pytest.raises(TypeError):
            rtr.append(1)
        with pytest.raises(TypeError):
            rtr.append("2323")

    def test_append_overlap(self):
        """
        Appending overlapping traces should raise a UserWarning/TypeError
        """
        rtr = RtTrace()
        tr = Trace(data=np.array([0, 1]))
        rtr.append(tr)
        # this raises UserWarning
        with warnings.catch_warnings(record=True):
            warnings.simplefilter('error', UserWarning)
            with pytest.raises(UserWarning):
                rtr.append(tr)
        # append with gap_overlap_check=True will raise a TypeError
        with pytest.raises(TypeError):
            rtr.append(tr, gap_overlap_check=True)

    def test_append_gap(self):
        """
        Appending a traces with a time gap should raise a UserWarning/TypeError
        """
        rtr = RtTrace()
        tr = Trace(data=np.array([0, 1]))
        tr2 = Trace(data=np.array([5, 6]))
        tr2.stats.starttime = tr.stats.starttime + 10
        rtr.append(tr)
        # this raises UserWarning
        with warnings.catch_warnings(record=True):
            warnings.simplefilter('error', UserWarning)
            with pytest.raises(UserWarning):
                rtr.append(tr2)
        # append with gap_overlap_check=True will raise a TypeError
        with pytest.raises(TypeError):
            rtr.append(tr2, gap_overlap_check=True)

    def test_copy(self):
        """
        Testing copy of RtTrace object.
        """
        rtr = RtTrace()
        rtr.copy()
        # register predefined function
        rtr.register_rt_process('integrate', test=1, muh='maeh')
        rtr.copy()
        # register ObsPy function call
        rtr.register_rt_process(obspy.signal.filter.bandpass, freqmin=0,
                                freqmax=1, df=0.1)
        rtr.copy()
        # register NumPy function call
        rtr.register_rt_process(np.square)
        rtr.copy()

    def test_append_not_float32(self):
        """
        Test for not using float32.
        """
        tr = read()[0]
        tr.data = np.require(tr.data, dtype='>f4')
        traces = tr / 3
        rtr = RtTrace()
        for trace in traces:
            rtr.append(trace)

    def test_missing_or_wrong_argument_in_rt_process(self):
        """
        Tests handling of missing/wrong arguments.
        """
        trace = Trace(np.arange(100))
        # 1- function scale needs no additional arguments
        rt_trace = RtTrace()
        rt_trace.register_rt_process('scale')
        rt_trace.append(trace)
        # adding arbitrary arguments should fail
        rt_trace = RtTrace()
        rt_trace.register_rt_process('scale', muh='maeh')
        with pytest.raises(TypeError):
            rt_trace.append(trace)
        # 2- function tauc has one required argument
        rt_trace = RtTrace()
        rt_trace.register_rt_process('tauc', width=10)
        rt_trace.append(trace)
        # wrong argument should fail
        rt_trace = RtTrace()
        rt_trace.register_rt_process('tauc', xyz='xyz')
        with pytest.raises(TypeError):
            rt_trace.append(trace)
        # missing argument width should raise an exception
        rt_trace = RtTrace()
        rt_trace.register_rt_process('tauc')
        with pytest.raises(TypeError):
            rt_trace.append(trace)
        # adding arbitrary arguments should fail
        rt_trace = RtTrace()
        rt_trace.register_rt_process('tauc', width=20, notexistingoption=True)
        with pytest.raises(TypeError):
            rt_trace.append(trace)



"""
obspy.io.csv - CSV, CSZ and EVENTTXT read/write support for earthquake catalogs
===============================================================================


Usage CSV
---------

CSV format can be used to store a catalog with basic origin properties.
Picks cannot be stored.

>>> from obspy import read_events
>>> events = read_events('/path/to/catalog.csv')
>>> print(events)  # doctest: +NORMALIZE_WHITESPACE
3 Event(s) in Catalog:
2012-04-04T14:21:42.300000Z | +41.818,  +79.689 | 4.4  mb
2012-04-04T14:18:37.000000Z | +39.342,  +41.044 | 4.3  ML
2012-04-04T14:08:46.000000Z | +38.017,  +37.736 | 3.0  ML
>>> events.write('local_catalog.csv', 'CSV')  # declare 'CSV' as format
>>> with open('local_catalog.csv') as f: print(f.read())
id,time,lat,lon,dep,magtype,mag
20120404_0000041,2012-04-04T14:21:42.30000,41.818000,79.689000,1.000,mb,4.40
20120404_0000038,2012-04-04T14:18:37.00000,39.342000,41.044000,14.400,ML,4.30
20120404_0000039,2012-04-04T14:08:46.00000,38.017000,37.736000,7.000,ML,3.00
<BLANKLINE>

It is possible to load arbitrary CSV files. Define the field names in the code
or use the first line in the file to define the field names.
The following field names have to be used to read the origin time:
`time` (UTC time string) or ``'year, mon, day, hour, minu, sec'``.
The following additional field names have to be used:
``lat, lon, dep, mag``. ``magtype``, ``id`` and some other fields are optional.
For external CSV files, the format ``'CSV'`` has to be explicitly specified.

>>> from obspy.core.util import get_example_file
>>> with open(get_example_file('external.csv')) as f: print(f.read())
Year, Month, Day, Hour, Minute, Seconds, code, Lat, Lon, Depth, Magnitude, ID
2023, 05, 06, 19, 55, 01.3, LI, 10.1942, 124.8300, 50.47, 0.2, 2023abcde
<BLANKLINE>
>>> names = 'year mon day hour minu sec _ lat lon dep mag id'
>>> events = read_events('/path/to/external.csv', 'CSV', skipheader=1, names=names)
>>> print(events)  # doctest: +NORMALIZE_WHITESPACE
1 Event(s) in Catalog:
2023-05-06T19:55:01.300000Z | +10.194, +124.830 | 0.2


Usage EVENTTXT
--------------

The EVENTTXT format is a flavour of CSV, reading and writing is directly
supported.

>>> from obspy import read_events
>>> print(read_events('/path/to/events.txt'))  # doctest: +NORMALIZE_WHITESPACE
2 Event(s) in Catalog:
2012-04-11T08:38:37.000000Z |  +2.238,  +93.014 | 8.6  MW
1960-05-22T19:11:14.000000Z | -38.170,  -72.570 | 8.5
>>> print(read_events('https://service.iris.edu/fdsnws/event/1/query?minmagnitude=8.5&format=text&endtime=2020-01-01'))  # doctest: +NORMALIZE_WHITESPACE
7 Event(s) in Catalog:
2012-04-11T08:38:37.000000Z |  +2.238,  +93.014 | 8.6  MW
2011-03-11T05:46:23.000000Z | +38.296, +142.498 | 9.1  MW
2010-02-27T06:34:13.000000Z | -36.148,  -72.933 | 8.8  MW
2007-09-12T11:10:26.000000Z |  -4.464, +101.396 | 8.5  MW
2005-03-28T16:09:35.000000Z |  +2.096,  +97.113 | 8.6  MW
2004-12-26T00:58:52.000000Z |  +3.413,  +95.901 | 9.0  MW
1960-05-22T19:11:14.000000Z | -38.170,  -72.570 | 8.5



Usage CSZ
---------

CSZ format can be used to store a catalog with picks in a set of csv files
zipped into a single file.
It works similar to NumPy's npz format.
Compression may be used with ``compression`` and ``compresslevel`` parameters
(see `zipfile doc <https://docs.python.org/library/zipfile.html#zipfile.ZipFile>`_).

>>> events = read_events('/path/to/example.pha')
>>> print(events)
2 Event(s) in Catalog:
2025-05-14T14:35:35.510000Z | +40.225,  +10.450 | 3.5  None
2025-05-14T15:43:05.280000Z | +40.223,  +10.450 | 1.8  None
>>> print(len(events[0].picks))
2
>>> events.write('catalog.csz', 'CSZ')
>>> events2 = read_events('catalog.csz')
>>> print(events2)  # doctest: +NORMALIZE_WHITESPACE
2 Event(s) in Catalog:
2025-05-14T14:35:35.510000Z | +40.225,  +10.450 | 3.5
2025-05-14T15:43:05.280000Z | +40.223,  +10.450 | 1.8
>>> print(len(events2[0].picks))
2

Load CSV/CSZ/EVENTTXT file into numpy array
-------------------------------------------

For plotting, e.t.c, it is useful to represent the event paramters with a numpy
array.
The :func:`load_csv` function can be used to load a CSV or CSZ file as numpy array.
The :func:`load_eventtxt` function ca be used to load an EVENTTXT file as numpy array.

>>> from obspy.io.csv import load_csv, load_eventtxt
>>> t = load_csv('/path/to/catalog.csv')
>>> print(t)  # doctest: +NORMALIZE_WHITESPACE
[ ('20120404_0000041', '2012-04-04T14:21:42.300',  41.818,  79.689,   1. , 'mb',  4.4)
  ('20120404_0000038', '2012-04-04T14:18:37.000',  39.342,  41.044,  14.4, 'ML',  4.3)
  ('20120404_0000039', '2012-04-04T14:08:46.000',  38.017,  37.736,   7. , 'ML',  3. )]
 >>> print(t['mag'])
 [ 4.4  4.3  3. ]
 >>> t2 = load_eventtxt('/path/to/events.txt')
 >>> print(t2)  # doctest: +NORMALIZE_WHITESPACE
 [ ('3337497', '2012-04-11T08:38:37.000', 2.2376, 93.0144,  26.3, 'MW',  8.6)
   ('2413', '1960-05-22T19:11:14.000', -38.17  , -72.57  ,   0. , '',  8.5)]

Convert ObsPy catalog into numpy array
--------------------------------------

The :func:`_events2array` function  can be used to convert an ObsPy catalog to numpy array.
Code example creating event plots::

    import matplotlib.pyplot as plt
    from obspy import read_events
    from obspy.io.csv import _events2array
    events = read_events()
    t = _events2array(events)
    plt.subplot(121)
    plt.scatter(t['lon'], t['lat'], 4*t['mag']**2)
    plt.subplot(122)
    plt.scatter(t['time'], t['mag'], 4*t['mag']**2)
    plt.show()
"""

from obspy.io.csv.core import _events2array, load_csv, load_eventtxt



import matplotlib.pyplot as plt
import cartopy.crs as ccrs

from obspy.imaging.beachball import beach


projection = ccrs.PlateCarree(central_longitude=142.0)

fig = plt.figure(dpi=150)
ax = fig.add_subplot(111, projection=projection)
ax.set_extent((-180, 180, -90, 90))
ax.coastlines()
ax.gridlines()

x, y = projection.transform_point(x=142.36929, y=38.3215,
                                  src_crs=ccrs.Geodetic())
focmecs = [0.136, -0.591, 0.455, -0.396, 0.046, -0.615]

ax = plt.gca()
b = beach(focmecs, xy=(x, y), width=10, linewidth=1, alpha=0.85)
b.set_zorder(10)
ax.add_collection(b)

plt.show()



# -*- coding: utf-8 -*-
#
# ObsPy documentation configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
#

import datetime
import obspy
import os
import sys

import cartopy  # NOQA  Do we really need this?

import matplotlib

# Use matplotlib agg backend
matplotlib.use("agg")


# -- Path setup --------------------------------------------------------------
# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
sys.path.append(os.path.abspath('_ext'))


# -- Project information -----------------------------------------------------
project = 'ObsPy'
author = 'The ObsPy Development Team (devs@obspy.org)'
year = datetime.date.today().year
copyright = '2012-{}, The ObsPy Development Team (devs@obspy.org)'.format(year)
version = ".".join(obspy.__version__.split(".")[:3])
release = obspy.__version__


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '3.1.1'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.intersphinx',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinx.ext.mathjax',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.duration',
    'sphinx.ext.todo',
    # theme
    'm2r2',
    'sphinx_rtd_theme',
    # custom extensions
    'plot_directive',
    'credits',
    'citations',
    # 'sphinxcontrib.bibtex',
]

# Uncomment this if you want to use sphinxcontrib.bibtex
# from pathlib import Path
# bibtex_bibfiles = (
#     list(Path('.').joinpath('source', 'bibliography').glob('*.bib'))
# )

# The file extensions of source files. Sphinx considers the files with this
# suffix as sources. The value can be a dictionary mapping file extensions to
# file types.
source_suffix = {
    '.rst': 'restructuredtext',
    '.md': 'markdown',
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The master toctree document.
master_doc = 'index'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    '_ext', '_images', '_static', '_templates', 'bibliography', 'credits']

# Warn about all references where the target cannot be found.
nitpicky = True
nitpick_ignore = [
    ('py:class', 'optional'),
    ('py:class', 'file'),
    ('py:class', 'file-like object'),
    ('py:class', 'open file'),
    ('py:class', 'valid matplotlib color'),
    ('py:class', 'valid matplotlib colormap'),
    ('py:class', 'same class as original object'),
    ('py:class', 'sqlalchemy.orm.decl_api.Base'),
    ('py:class', 'array_like'),
    ('py:class', 'hashable'),
    ('py:class', 'shapefile.Writer'),
    ('py:class', 'collections.namedtuple')
]

# suppress built-in types by default in nitpick
import builtins
for name in dir(builtins):
    nitpick_ignore += [('py:class', name)]

# suppress warnings for all Class built using the _event_type_class_factory
# which generate: "py:class reference target not found:
# obspy.core.event.base._event_type_class_factory.<locals>.
# AbstractEventTypeWithResourceID
nitpick_ignore_regex = [
    (r'py:class', r'.*<locals>\.AbstractEvent..*'),
    (r'py:class', r'math\..*'),
]

# configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference/', None),
    'matplotlib': ('https://matplotlib.org/', None),
    'sqlalchemy': ('https://docs.sqlalchemy.org/en/latest/', None),
    'pip': ('https://pip.pypa.io/en/stable/', None),
    'lxml': ('https://lxml.de/apidoc/', None),
    'cartopy': ('https://cartopy.readthedocs.io/stable/', None),
    'sphinx': ('https://www.sphinx-doc.org/en/master/', None),
    'sqlalchemy': ('https://docs.sqlalchemy.org/en/latest/', None),
    'requests': ('https://docs.python-requests.org/en/latest/', None)
}

# A boolean that decides whether module names are prepended to all object names
# (for object types where a “module” of some kind is defined).
add_module_names = False

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = ['obspy.']

# These values determine how to format the current date.
today_fmt = "%B %d %H o'clock, %Y"


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"

html_logo = '_static/obspy_logo_no_text.svg'
html_favicon = '_static/favicon.ico'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# These paths are either relative to html_static_path
# or fully qualified paths (eg. https://...)
html_css_files = [
    'css/custom.min.css',
]

# Additional templates that should be rendered to pages, maps page names to
# template names.
html_additional_pages = {
}

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%Y-%m-%dT%H:%M:%S'


# If true, a list all whose items consist of a single paragraph and/or a
# sub-list all whose items etc… (recursive definition) will not use the <p>
# element for any of its items.
html_compact_lists = True


# -- Options for LaTeX output -------------------------------------------------

# The paper size ('letter' or 'a4').
# latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
# latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual])
latex_documents = [
  ('tutorial/index', 'ObsPyTutorial.tex', u'ObsPy Tutorial',
   u'The ObsPy Development Team (devs@obspy.org)', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Additional stuff for the LaTeX preamble.
# latex_preamble = ''

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output -------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'obspydocumentation', u'ObsPy Documentation',
     [u'The ObsPy Development Team (devs@obspy.org)'], 1)
]


# -- Options for autodoc / autosummary exensions -----------------------------

# Don't merge __init__ method in auoclass content
autoclass_content = 'class'

# generate automatically stubs
autosummary_generate = True

# If true, autosummary already overwrites stub files by generated contents.
autosummary_generate_overwrite = False

# Not sure this will remove warnings from collections's Mapping (Attribdict)
autodoc_inherit_docstrings = False

# -- Options for linkcheck exension ------------------------------------------
linkcheck_timeout = 5
linkcheck_workers = 10


# -- Options for matplotlib plot directive -----------------------------------
# File formats to generate.
plot_formats = [('png', 110), ]  # ('hires.png', 200), ('pdf', 200)]



# -*- coding: utf-8 -*-
"""
JSON Encoder default function

This module provides:
---------------------
Default : a class to create a "default" function accepted by the
python :py:mod:`json` module Encoder classes, valid for
:class:`~obspy.core.event.event.Event` objects.

Example
-------
>>> import json
>>> from obspy import read_events
>>> from obspy.io.json.default import Default
>>> c = read_events()
>>> d = Default(omit_nulls=False)
>>> s = json.dumps(c, default=d)

"""
from obspy import UTCDateTime
from obspy.core.event import Catalog, ResourceIdentifier
from obspy.core.util import AttribDict


class Default(object):
    """
    Class to create a "default" function for the
    :py:func:`json.dump`/:py:func:`json.dumps` functions
    which is passed to the JSONEncoder.

    """
    _catalog_attrib = ('events', 'comments', 'description', 'creation_info',
                       'resource_id')

    OMIT_NULLS = None
    TIME_FORMAT = None

    def __init__(self, omit_nulls=True, time_format=None):
        """
        Create a "default" function for JSONEncoder for ObsPy objects

        :param bool omit_nulls: Leave out any null or empty values (True)
        :param str time_format: Format string passed to strftime (None)

        """
        # Allows customization of the function
        self.OMIT_NULLS = omit_nulls
        self.TIME_FORMAT = time_format

    def __call__(self, obj):
        """
        Deal with :class:`~obspy.core.event.event.Event` objects in JSON
        Encoder

        This function can be passed to the json module's
        `default` keyword parameter

        """
        # Most event objects have dict methods, construct a dict
        # and deal with special cases that don't
        if isinstance(obj, AttribDict):
            # Map to a serializable dict
            # Leave out nulls, empty strings, list, dicts, except for numbers
            if self.OMIT_NULLS:
                return {k: v for k, v in obj.items() if v or v == 0}
            else:
                return {k: v for k, v in obj.items()}
        elif isinstance(obj, Catalog):
            # Catalog isn't a dict
            return {k: getattr(obj, k)
                    for k in self._catalog_attrib if getattr(obj, k)}
        elif isinstance(obj, UTCDateTime):
            if self.TIME_FORMAT is None:
                return str(obj)
            else:
                return obj.strftime(self.TIME_FORMAT)
        elif isinstance(obj, ResourceIdentifier):
            # Always want ID as a string
            return str(obj)
        else:
            return None



import obspy


st = obspy.read("https://examples.obspy.org/GR.BFO..LHZ.2012.108")
st.filter("lowpass", freq=0.1, corners=2)
st.plot(type="dayplot", interval=60, right_vertical_labels=False,
        vertical_scaling_range=5e3, one_tick_per_line=True,
        color=['k', 'r', 'b', 'g'], show_y_UTC_label=False,
        events={'min_magnitude': 6.5})



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import io
import re
import warnings

import numpy as np

import obspy
from obspy.core.util import NamedTemporaryFile, CatchAndAssertWarnings
from obspy.io.reftek.core import (
    _read_reftek130, _is_reftek130, Reftek130, Reftek130Exception)
from obspy.io.reftek.packet import (
    _unpack_C0_C2_data_fast, _unpack_C0_C2_data_safe, _unpack_C0_C2_data,
    EHPacket, _initial_unpack_packets)
import pytest


class TestReftek():
    """
    Test suite for obspy.io.reftek
    """
    @pytest.fixture(autouse=True, scope="function")
    def setup(self, testdata):
        try:
            # doctests of __init__.py produce warnings that get caught. if we
            # don't raze the slate out the registry here, we can't test those
            # warnings in the unit tests (if doctests run before unittests)..
            from obspy.io.reftek.core import __warningregistry__
            __warningregistry__.clear()
        except ImportError:
            # import error means no warning has been issued
            # before, so nothing to do.
            pass
        self.reftek_filename = "225051000_00008656"
        self.reftek_file = testdata[self.reftek_filename]
        self.reftek_file_steim2 = testdata['104800000_000093F8']
        self.reftek_file_16 = testdata['065520000_013EE8A0.rt130']
        self.reftek_file_16_npz = testdata['065520000_013EE8A0.npz']
        self.reftek_file_32 = testdata['230000005_0036EE80_cropped.rt130']
        self.reftek_file_32_npz = testdata['230000005_0036EE80_cropped.npz']
        self.mseed_filenames = [
            "2015282_225051_0ae4c_1_1.msd",
            "2015282_225051_0ae4c_1_2.msd", "2015282_225051_0ae4c_1_3.msd"]
        self.mseed_files = [
            testdata[filename] for filename in self.mseed_filenames]
        # files "2015282_225051_0ae4c_1_[123].msd" contain miniseed data
        # converted with "rt_mseed" tool of Reftek utilities.

        # information on the data packets in the test file:
        # (omitting leading/trailing EH/ET packet)
        #   >>> for p in packets[1:-1]:
        #   ...     print(
        #   ...         "{:02d}".format(p.packet_sequence),
        #   ...         p.channel_number,
        #   ...         "{:03d}".format(p.number_of_samples),
        #   ...         p.time)
        #   01 0 549 2015-10-09T22:50:51.000000Z
        #   02 1 447 2015-10-09T22:50:51.000000Z
        #   03 2 805 2015-10-09T22:50:51.000000Z
        #   04 0 876 2015-10-09T22:50:53.745000Z
        #   05 1 482 2015-10-09T22:50:53.235000Z
        #   06 1 618 2015-10-09T22:50:55.645000Z
        #   07 2 872 2015-10-09T22:50:55.025000Z
        #   08 0 892 2015-10-09T22:50:58.125000Z
        #   09 1 770 2015-10-09T22:50:58.735000Z
        #   10 2 884 2015-10-09T22:50:59.385000Z
        #   11 0 848 2015-10-09T22:51:02.585000Z
        #   12 1 790 2015-10-09T22:51:02.585000Z
        #   13 2 844 2015-10-09T22:51:03.805000Z
        #   14 0 892 2015-10-09T22:51:06.215000Z
        #   15 1 768 2015-10-09T22:51:05.925000Z
        #   16 2 884 2015-10-09T22:51:08.415000Z
        #   17 1 778 2015-10-09T22:51:10.765000Z
        #   18 0 892 2015-10-09T22:51:11.675000Z
        #   19 2 892 2015-10-09T22:51:12.835000Z
        #   20 1 736 2015-10-09T22:51:14.655000Z
        #   21 0 892 2015-10-09T22:51:16.135000Z
        #   22 2 860 2015-10-09T22:51:17.295000Z
        #   23 1 738 2015-10-09T22:51:18.335000Z
        #   24 0 892 2015-10-09T22:51:20.595000Z
        #   25 1 673 2015-10-09T22:51:22.025000Z
        #   26 2 759 2015-10-09T22:51:21.595000Z
        #   27 0 067 2015-10-09T22:51:25.055000Z
        self.reftek_file_vpu = testdata['221935615_00000000']

    def _assert_reftek130_test_stream(self, st_reftek):
        """
        Test reftek 130 data read into a stream object against miniseed files
        converted using "rt_mseed" utility from Trimble/Reftek.

        Note that rt_mseed fills in network as "XX", location as "01" and
        channels as "001", "002", "003".
        """
        st_mseed = obspy.Stream()
        for file_ in self.mseed_files:
            st_mseed += obspy.read(file_, "MSEED")
        # reftek reader correctly fills in band+instrument code but rt_mseed
        # does not apparently, so set it now for the comparison
        for tr in st_mseed:
            tr.stats.channel = "EH" + tr.stats.channel[-1]
            tr.stats.pop("_format")
            tr.stats.pop("mseed")
        # check reftek130 low-level headers separately:
        for tr in st_reftek:
            assert "reftek130" in tr.stats
            # XXX TODO check reftek specific headers
            tr.stats.pop("reftek130")
            tr.stats.pop("_format", None)
        # sort streams
        st_reftek = st_reftek.sort()
        st_mseed = st_mseed.sort()
        # check amount of traces
        assert len(st_reftek) == len(st_mseed)
        # check equality of headers
        for tr_got, tr_expected in zip(st_reftek, st_mseed):
            assert tr_got.stats == tr_expected.stats
        # check equality of data
        for tr_got, tr_expected in zip(st_reftek, st_mseed):
            np.testing.assert_array_equal(tr_got.data, tr_expected.data)

    def test_read_reftek130_steim1(self):
        """
        Test original reftek 130 data file against miniseed files converted
        using "rt_mseed" utility from Trimble/Reftek.

        rt_mseed fills in network as "XX", location as "01" and channels as
        "001", "002", "003".
        """
        st_reftek = _read_reftek130(
            self.reftek_file, network="XX", location="01",
            component_codes=["1", "2", "3"],
            sort_permuted_package_sequence=True)
        self._assert_reftek130_test_stream(st_reftek)

    def test_read_reftek130_steim2(self):
        """
        Test reading a steim2 encoded data file.

        Unpacking of data is tested separately so just checking a few samples
        at the start should suffice.
        """
        st = _read_reftek130(
            self.reftek_file_steim2, network="XX", location="01",
            component_codes=["1", "2", "3"],
            sort_permuted_package_sequence=True)
        # note: test data has stream name defined as 'DS 1', so we end up with
        # non-SEED conforming channel codes which is expected
        assert len(st) == 3
        assert len(st[0]) == 3788
        assert len(st[1]) == 3788
        assert len(st[2]) == 3788
        assert st[0].id == 'XX.TL01.01.DS 11'
        assert st[1].id == 'XX.TL01.01.DS 12'
        assert st[2].id == 'XX.TL01.01.DS 13'
        np.testing.assert_array_equal(
            st[0].data[:5], [26814, 26823, 26878, 26941, 26942])
        np.testing.assert_array_equal(
            st[1].data[:5], [-1987, -1984, -1959, -1966, -1978])
        np.testing.assert_array_equal(
            st[2].data[:5], [-2404, -2376, -2427, -2452, -2452])

    def test_read_reftek130_no_component_codes_specified(self):
        """
        Test reading reftek 130 data file not providing component codes
        (relying on the information in header packet).

        rt_mseed fills in network as "XX", location as "01" and channels as
        "001", "002", "003".
        """
        msg = ('No channel code specified in the data file and no component '
               'codes specified. Using stream label and number of channel '
               'in file as channel codes.')
        with CatchAndAssertWarnings(expected=[(UserWarning, msg)]) as w:
            warnings.simplefilter("always")
            st_reftek = _read_reftek130(
                self.reftek_file, network="XX", location="01",
                sort_permuted_package_sequence=True)
        assert len(w) >= 8
        # check that channel codes are set with stream label from EH packet +
        # enumerated channel number starting at 0
        for tr, cha in zip(st_reftek, ('EH0', 'EH0', 'EH0', 'EH1', 'EH1',
                                       'EH1', 'EH2', 'EH2')):
            assert tr.stats.channel == cha
        for tr in st_reftek:
            # need to adapt channel codes to compare against mseed stream now..
            tr.stats.channel = (
                'EH' + {'0': '1', '1': '2', '2': '3'}[tr.stats.channel[-1]])
        self._assert_reftek130_test_stream(st_reftek)

    def test_is_reftek130(self):
        """
        Test checking whether file is REFTEK130 format or not.
        """
        assert _is_reftek130(self.reftek_file)
        for file_ in self.mseed_files:
            assert not _is_reftek130(file_)

    def test_integration_with_obspy_core(self):
        """
        Test the integration with ObsPy core.
        """
        st_reftek = obspy.read(
            self.reftek_file, network="XX", location="01",
            component_codes=["1", "2", "3"])
        self._assert_reftek130_test_stream(st_reftek)

    def test_error_no_packets_read(self):
        """
        Test error message when no packets could be read from file.
        """
        with NamedTemporaryFile() as fh:
            # try to read empty file, finding no packets
            with pytest.raises(Reftek130Exception):
                _read_reftek130(fh.name)
        # try to read mseed file, finding no packets
        with pytest.raises(Reftek130Exception):
            _read_reftek130(self.mseed_files[0],
                            sort_permuted_package_sequence=True)

    def test_warning_disturbed_packet_sequence(self):
        """
        Test warning message when packet sequence is non-contiguous (one packet
        missing).
        """
        with NamedTemporaryFile() as fh:
            with open(self.reftek_file, 'rb') as fh2:
                # write packages to the file and omit one packet
                # (packets are 1024 byte each)
                fh.write(fh2.read(1024 * 5))
                fh2.seek(1024 * 6)
                fh.write(fh2.read())
            fh.seek(0)
            # try to read file, finding a non-contiguous packet sequence
            msg = 'Detected a non-contiguous packet sequence!'
            with CatchAndAssertWarnings(expected=[(UserWarning, msg)]):
                warnings.simplefilter("always")
                _read_reftek130(fh.name, network="XX", location="01",
                                component_codes=["1", "2", "3"],
                                sort_permuted_package_sequence=True)

    def test_read_file_perturbed_packet_sequence(self):
        """
        Test data read from file when packet sequence is perturbed. This makes
        sure that data is read correctly even when the array storing the
        packets gets permuted and thus becomes incontiguous for C.
        """
        with NamedTemporaryFile() as fh:
            with open(self.reftek_file, 'rb') as fh2:
                # write packages to the file and move some packets around in
                # the file
                # (packets are 1024 byte each)
                tmp1 = fh2.read(1024 * 2)
                tmp2 = fh2.read(1024 * 4)
                tmp3 = fh2.read(1024 * 1)
                tmp4 = fh2.read(1024 * 2)
                tmp5 = fh2.read(1024 * 3)
                tmp6 = fh2.read()
            fh.write(tmp1)
            fh.write(tmp3)
            fh.write(tmp5)
            fh.write(tmp6)
            fh.write(tmp2)
            fh.write(tmp4)
            fh.seek(0)
            # try to read file, finding a non-contiguous packet sequence
            msg = 'Detected permuted packet sequence, sorting.'
            with CatchAndAssertWarnings(expected=[(UserWarning, msg)]):
                warnings.simplefilter("always")
                st_reftek = _read_reftek130(
                    fh.name, network="XX", location="01",
                    component_codes=["1", "2", "3"],
                    sort_permuted_package_sequence=True)
        st_reftek.merge(-1)
        self._assert_reftek130_test_stream(st_reftek)

    def test_drop_not_implemented_packets(self):
        """
        Test error message when some not implemented packet types are dropped
        """
        with NamedTemporaryFile() as fh:
            with open(self.reftek_file, 'rb') as fh2:
                # write packages to the file and write the last three packets
                # with a different packet type
                # (packets are 1024 byte each)
                tmp = fh2.read()[:-(1024 * 3)]
                fh2.seek(-(1024 * 3), 2)
                tmp2 = fh2.read(1024)
                tmp3 = fh2.read(1024)
                tmp4 = fh2.read(1024)
            # write last three packages with some different packet types
            fh.write(tmp)
            fh.write(b"AA")
            fh.write(tmp2[2:])
            fh.write(b"AA")
            fh.write(tmp3[2:])
            fh.write(b"BB")
            fh.write(tmp4[2:])
            fh.seek(0)
            msg = re.escape(
                "Encountered some packets of types that are not "
                "implemented yet (types: [b'AA', b'BB']). Dropped 3 "
                "packets overall.")
            # this message we get because ET packet at end is now missing
            msg2 = re.escape(
                'No event trailer (ET) packets in packet sequence. File '
                'might be truncated.')
            with CatchAndAssertWarnings(
                    expected=[(UserWarning, msg), (UserWarning, msg2)]) as w:
                warnings.simplefilter("always")
                _read_reftek130(
                    fh.name, network="XX", location="01",
                    component_codes=["1", "2", "3"],
                    sort_permuted_package_sequence=True)
            assert len(w) >= 2

    def test_missing_event_trailer_packet(self):
        """
        Test that reading the file if the ET packet is missing works and a
        warning is shown.
        """
        with NamedTemporaryFile() as fh:
            with open(self.reftek_file, 'rb') as fh2:
                # write packages to the file and omit last (ET) packet
                # (packets are 1024 byte each)
                tmp = fh2.read()
                tmp = tmp[:-1024]
            fh.write(tmp)
            fh.seek(0)
            msg = re.escape(
                'No event trailer (ET) packets in packet sequence. File '
                'might be truncated.')
            with CatchAndAssertWarnings(expected=[(UserWarning, msg)]):
                warnings.simplefilter("always")
                st_reftek = _read_reftek130(
                    fh.name, network="XX", location="01",
                    component_codes=["1", "2", "3"],
                    sort_permuted_package_sequence=True)
        self._assert_reftek130_test_stream(st_reftek)

    def test_truncated_last_packet(self):
        """
        Test that reading the file works, if the ET packet at the end is
        truncated and thus omitted.
        """
        with NamedTemporaryFile() as fh:
            with open(self.reftek_file, 'rb') as fh2:
                # write packages to the file and truncate last (ET) packet
                # (packets are 1024 byte each)
                tmp = fh2.read()
                tmp = tmp[:-10]
            fh.write(tmp)
            fh.seek(0)
            # we get two warnings, one about the truncated packet and one about
            # the missing last (ET) packet
            msg = re.escape(
                'Length of data not a multiple of '
                '1024. Data might be truncated. Dropping 1014 '
                'byte(s) at the end.')
            msg2 = re.escape(
                'No event trailer (ET) packets in packet sequence. File '
                'might be truncated.')
            with CatchAndAssertWarnings(
                    expected=[(UserWarning, msg), (UserWarning, msg2)]) as w:
                warnings.simplefilter("always")
                st_reftek = _read_reftek130(
                    fh.name, network="XX", location="01",
                    component_codes=["1", "2", "3"],
                    sort_permuted_package_sequence=True)
            assert len(w) >= 2
        # data should be read OK aside from the warnings
        self._assert_reftek130_test_stream(st_reftek)

    def test_no_eh_et_packet(self):
        """
        Test error messages when reading a file without any EH/ET packet.
        """
        msg = \
            "Reftek data contains data packets without corresponding header " \
            "or trailer packet."
        with NamedTemporaryFile() as fh:
            with open(self.reftek_file, 'rb') as fh2:
                # write packages to the file and omit first and last (EH/ET)
                # packet
                # (packets are 1024 byte each)
                tmp = fh2.read()
            fh.write(tmp[1024:-1024])
            fh.seek(0)
            with pytest.raises(Reftek130Exception, match=msg):
                _read_reftek130(
                    fh.name, network="XX", location="01",
                    component_codes=["1", "2", "3"],
                    sort_permuted_package_sequence=True)

    def test_data_unpacking_steim1(self, testdata):
        """
        Test both unpacking routines for C0 data coding (STEIM1)
        """
        rt = Reftek130.from_file(self.reftek_file)
        expected = np.load(testdata['unpacked_data_steim1.npy'])
        packets = rt._data[rt._data['packet_type'] == b'DT'][:10]
        for func in (_unpack_C0_C2_data, _unpack_C0_C2_data_fast,
                     _unpack_C0_C2_data_safe):
            got = func(packets, encoding='C0')
            np.testing.assert_array_equal(got, expected)

    def test_data_unpacking_steim2(self, testdata):
        """
        Test both unpacking routines for C2 data coding (STEIM2)
        """
        rt = Reftek130.from_file(self.reftek_file_steim2)
        expected = np.load(testdata['unpacked_data_steim2.npy'])
        packets = rt._data[rt._data['packet_type'] == b'DT'][:10]
        for func in (_unpack_C0_C2_data, _unpack_C0_C2_data_fast,
                     _unpack_C0_C2_data_safe):
            got = func(packets, encoding='C2')
            np.testing.assert_array_equal(got, expected)

    def test_string_representations(self):
        """
        Test string representations of Reftek object and Packets
        """
        expected = [
            "Reftek130 (29 packets, file: {})".format(self.reftek_file),
            "Packet Sequence  Byte Count  Data Fmt  Sampling Rate      Time",
            "  | Packet Type   |  Event #  | Station | Channel #         |",
            "  |   |  Unit ID  |    | Data Stream #  |   |  # of samples |",
            "  |   |   |  Exper.#   |   |  |  |      |   |    |          |",
            "0000 EH AE4C  0  416  427  0 C0 KW1    200.        "
            "2015-10-09T22:50:51.000000Z",
            "0001 DT AE4C  0 1024  427  0 C0             0  549 "
            "2015-10-09T22:50:51.000000Z",
            "0002 DT AE4C  0 1024  427  0 C0             1  447 "
            "2015-10-09T22:50:51.000000Z",
            "0003 DT AE4C  0 1024  427  0 C0             2  805 "
            "2015-10-09T22:50:51.000000Z",
            "0004 DT AE4C  0 1024  427  0 C0             0  876 "
            "2015-10-09T22:50:53.745000Z",
            "0005 DT AE4C  0 1024  427  0 C0             1  482 "
            "2015-10-09T22:50:53.235000Z",
            "0006 DT AE4C  0 1024  427  0 C0             1  618 "
            "2015-10-09T22:50:55.645000Z",
            "0007 DT AE4C  0 1024  427  0 C0             2  872 "
            "2015-10-09T22:50:55.025000Z",
            "0008 DT AE4C  0 1024  427  0 C0             0  892 "
            "2015-10-09T22:50:58.125000Z",
            "0009 DT AE4C  0 1024  427  0 C0             1  770 "
            "2015-10-09T22:50:58.735000Z",
            "0010 DT AE4C  0 1024  427  0 C0             2  884 "
            "2015-10-09T22:50:59.385000Z",
            "0011 DT AE4C  0 1024  427  0 C0             0  848 "
            "2015-10-09T22:51:02.585000Z",
            "0012 DT AE4C  0 1024  427  0 C0             1  790 "
            "2015-10-09T22:51:02.585000Z",
            "0013 DT AE4C  0 1024  427  0 C0             2  844 "
            "2015-10-09T22:51:03.805000Z",
            "0014 DT AE4C  0 1024  427  0 C0             0  892 "
            "2015-10-09T22:51:06.215000Z",
            "0015 DT AE4C  0 1024  427  0 C0             1  768 "
            "2015-10-09T22:51:05.925000Z",
            "0016 DT AE4C  0 1024  427  0 C0             2  884 "
            "2015-10-09T22:51:08.415000Z",
            "0017 DT AE4C  0 1024  427  0 C0             1  778 "
            "2015-10-09T22:51:10.765000Z",
            "0018 DT AE4C  0 1024  427  0 C0             0  892 "
            "2015-10-09T22:51:11.675000Z",
            "0019 DT AE4C  0 1024  427  0 C0             2  892 "
            "2015-10-09T22:51:12.835000Z",
            "0020 DT AE4C  0 1024  427  0 C0             1  736 "
            "2015-10-09T22:51:14.655000Z",
            "0021 DT AE4C  0 1024  427  0 C0             0  892 "
            "2015-10-09T22:51:16.135000Z",
            "0022 DT AE4C  0 1024  427  0 C0             2  860 "
            "2015-10-09T22:51:17.295000Z",
            "0023 DT AE4C  0 1024  427  0 C0             1  738 "
            "2015-10-09T22:51:18.335000Z",
            "0024 DT AE4C  0 1024  427  0 C0             0  892 "
            "2015-10-09T22:51:20.595000Z",
            "0025 DT AE4C  0 1024  427  0 C0             1  673 "
            "2015-10-09T22:51:22.025000Z",
            "0026 DT AE4C  0 1024  427  0 C0             2  759 "
            "2015-10-09T22:51:21.595000Z",
            "0027 DT AE4C  0 1024  427  0 C0             0   67 "
            "2015-10-09T22:51:25.055000Z",
            "0028 ET AE4C  0  416  427  0 C0 KW1    200.        "
            "2015-10-09T22:50:51.000000Z",
            "(detailed packet information with: "
            "'print(Reftek130.__str__(compact=False))')"]
        with warnings.catch_warnings(record=True):
            warnings.simplefilter("always")
            rt130 = Reftek130.from_file(self.reftek_file)
        assert expected == str(rt130).splitlines()

    def test_reading_packet_with_vpu_float_string(self):
        """
        Test reading a data stream with VPU floating point in header, see #1632
        """
        with open(self.reftek_file_vpu, 'rb') as fh:
            data = fh.read(1024)
        data = _initial_unpack_packets(data)
        eh = EHPacket(data[0])
        assert eh.channel_sensor_vpu == \
            (2.4, 2.4, 2.4, None, None, None, None, None, None, None, None,
             None, None, None, None, None)
        # reading the file should work..
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            st = obspy.read(self.reftek_file_vpu,
                            sort_permuted_package_sequence=True)
        assert len(st) == 2
        assert len(st[0]) == 890
        assert len(st[1]) == 890
        np.testing.assert_array_equal(
            st[0][:10], [210, 212, 208, 211, 211, 220, 216, 215, 219, 218])

    def test_reading_file_with_multiple_events(self):
        """
        Test reading a "multiplexed" file with multiple "events" in it.

        Simply reuse the existing test data in one read operation.
        """
        with open(self.reftek_file_vpu, 'rb') as fh:
            data = fh.read()
        with open(self.reftek_file, 'rb') as fh:
            data += fh.read()
        bytes_ = io.BytesIO(data)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            st = obspy.read(bytes_, format='REFTEK130',
                            sort_permuted_package_sequence=True)
        assert len(st) == 10
        # we should have data from two different files/stations in there
        for tr in st[:8]:
            assert tr.stats.station == 'KW1'
        for tr in st[8:]:
            assert tr.stats.station == 'TL02'

    def test_reading_file_with_no_data_in_channel_zero(self):
        """
        Test reading a file that has no data packets in channel zero (e.g.
        6-channel Reftek and only recording on channels 4-6)

        Simply reuse the existing test data omitting the data packet that has
        channel zero.
        """
        with open(self.reftek_file_vpu, 'rb') as fh:
            data = fh.read()
        # only use first packet (the EH packet) and last packet (a DT packet
        # for channel number 1, i.e. channel 2)
        data = data[:1024] + data[-1024:]
        bio = io.BytesIO(data)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            st = obspy.read(bio, format='REFTEK130')
        assert len(st) == 1
        # just a few basic checks, reading data is covered in other tests
        tr = st[0]
        assert tr.id == ".TL02..DS 11"
        assert len(tr) == 890

    def test_reading_file_with_encoding_32(self):
        """
        Test reading a file with encoding '32' (uncompressed 32 bit integer)

        Only tests unpacked sample data, everything else should be covered by
        existing tests.
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            st = _read_reftek130(self.reftek_file_32)
        # read expected data
        npz = np.load(self.reftek_file_32_npz)
        # compare it
        assert len(st) == 3
        for tr, (_, expected) in zip(st, sorted(npz.items())):
            np.testing.assert_array_equal(expected, tr.data)

    def test_reading_file_with_encoding_16(self):
        """
        Test reading a file with encoding '16' (uncompressed 16 bit integer)

        Only tests unpacked sample data, everything else should be covered by
        existing tests.
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            st = _read_reftek130(self.reftek_file_16)
        # read expected data
        npz = np.load(self.reftek_file_16_npz)
        # compare it
        assert len(st) == 3
        assert len(st[0]) == 2090
        assert len(st[1]) == 2090
        assert len(st[2]) == 2090
        for tr, (_, expected) in zip(st, sorted(npz.items())):
            np.testing.assert_array_equal(expected, tr.data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Base class for all FDSN routers.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
    Celso G Reyes, 2017
    EarthScope (former IRIS-DMC)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from multiprocessing.dummy import Pool as ThreadPool

import decorator
import io
import sys
import traceback
import warnings
from urllib.parse import urlparse

import obspy
from obspy.core.util.deprecation_helpers import ObsPyDeprecationWarning

from ...base import HTTPClient
from .. import client
from ..client import raise_on_error
from ..header import FDSNException, URL_MAPPINGS, FDSNNoDataException


def RoutingClient(routing_type, *args, **kwargs):  # NOQA
    """
    Helper function to get the correct routing instance.

    :type routing_type: str
    :param routing_type: The type of router to initialize.
        ``"earthscope-federator"`` or ``"eida-routing"``. Will consequently
        return either a
        :class:`~.federator_routing_client.FederatorRoutingClient` or a
        :class:`~.eidaws_routing_client.EIDAWSRoutingClient` object,
        respectively.

    Remaining ``args`` and ``kwargs`` will be passed to the underlying classes.
    For example, credentials can be supported for all underlying data centers.
    See :meth:`BaseRoutingClient <BaseRoutingClient.__init__>` for details.

    >>> from obspy.clients.fdsn import RoutingClient

    Get an instance of a routing client using the EarthScope (former IRIS)
    Federator:

    >>> c = RoutingClient("earthscope-federator")
    >>> print(type(c))  # doctest: +ELLIPSIS
    <class '...routing.federator_routing_client.FederatorRoutingClient'>

    Or get an instance of a routing client using the EIDAWS routing web
    service:

    >>> c = RoutingClient("eida-routing")
    >>> print(type(c))  # doctest: +ELLIPSIS
    <class '...routing.eidaws_routing_client.EIDAWSRoutingClient'>
    """
    if routing_type.lower() == "iris-federator":
        routing_type = "earthscope-federator"
        msg = ("IRIS is now EarthScope, please consider changing the "
               "'routing_type' to 'earthscope-federator'.")
        warnings.warn(msg, ObsPyDeprecationWarning)

    if routing_type.lower() == "eida-routing":
        from .eidaws_routing_client import EIDAWSRoutingClient
        return EIDAWSRoutingClient(*args, **kwargs)
    elif routing_type.lower() == "earthscope-federator":
        from .federator_routing_client import FederatorRoutingClient
        return FederatorRoutingClient(*args, **kwargs)
    else:
        raise NotImplementedError(
            "Routing type '%s' is not implemented. Available types: "
            "`earthscope-federator`, `eida-routing`" % routing_type)


@decorator.decorator
def _assert_format_not_in_kwargs(f, *args, **kwargs):
    if "format" in kwargs:
        raise ValueError("The `format` argument is not supported")
    return f(*args, **kwargs)


@decorator.decorator
def _assert_filename_not_in_kwargs(f, *args, **kwargs):
    if "filename" in kwargs:
        raise ValueError("The `filename` argument is not supported")
    return f(*args, **kwargs)


@decorator.decorator
def _assert_attach_response_not_in_kwargs(f, *args, **kwargs):
    if "attach_response" in kwargs:
        raise ValueError("The `attach_response` argument is not supported")
    return f(*args, **kwargs)


def _try_download_bulk(r):
    try:
        return _download_bulk(r)
    except Exception:
        reason = "".join(traceback.format_exception(*sys.exc_info()))
        warnings.warn(
            "Failed to download data of type '%s' from '%s' due to: \n%s" % (
                r["data_type"], r["endpoint"], reason))
        return None


def _download_bulk(r):
    # Figure out the passed credentials, if any. Two possibilities:
    # (1) User and password, given explicitly for the base URLs (or an
    #     explicity given `eida_token` key per URL).
    # (2) A global EIDA_TOKEN key. It will be used for all services that
    #     don't have explicit credentials and also support the `/auth` route.
    credentials = r["credentials"].get(urlparse(r["endpoint"]).netloc, {})
    try:
        c = client.Client(r["endpoint"], debug=r["debug"],
                          timeout=r["timeout"], **credentials)
    # This should rarely happen but better safe than sorry.
    except FDSNException as e:  # pragma: no cover
        msg = e.args[0]
        msg += "It will not be used for routing. Try again later?"
        warnings.warn(msg)
        return None

    if not credentials and "EIDA_TOKEN" in r["credentials"] and \
            c._has_eida_auth:
        c.set_eida_token(r["credentials"]["EIDA_TOKEN"])

    if r["data_type"] == "waveform":
        fct = c.get_waveforms_bulk
        service = c.services["dataselect"]
    elif r["data_type"] == "station":
        fct = c.get_stations_bulk
        service = c.services["station"]

    # Keep only kwargs that are supported by this particular service.
    kwargs = {k: v for k, v in r["kwargs"].items() if k in service}
    bulk_str = ""
    for key, value in kwargs.items():
        bulk_str += "%s=%s\n" % (key, str(value))
    try:
        return fct(bulk_str + r["bulk_str"])
    except FDSNException:
        return None


def _strip_protocol(url):
    url = urlparse(url)
    return url.netloc + url.path


# Does not inherit from the FDSN client as that would be fairly hacky as
# some methods just make no sense for the routing client to have (e.g.
# get_events() but also others).
class BaseRoutingClient(HTTPClient):
    def __init__(self, debug=False, timeout=120, include_providers=None,
                 exclude_providers=None, credentials=None):
        """
        :type routing_type: str
        :param routing_type: The type of
            router to initialize. For details see :func:`RoutingClient`.
        :type exclude_providers: str or list[str]
        :param exclude_providers: Get no data from these providers. Can be
            the full HTTP address or one of the shortcuts ObsPy knows about.
        :type include_providers: str or list[str]
        :param include_providers: Get data only from these providers. Can be
            the full HTTP address of one of the shortcuts ObsPy knows about.
        :type credentials: dict
        :param credentials: Credentials for the individual data centers as a
            dictionary that maps base url of FDSN web service to either
            username/password or EIDA token, e.g.
            ``credentials={
            'geofon.gfz.de': {'eida_token': 'my_token_file.txt'},
            'service.earthscope.org': {'user': 'me', 'password': 'my_pass'}
            'EIDA_TOKEN': '/path/to/token.txt'
            }``
            The root level ``'EIDA_TOKEN'`` will be applied to all data centers
            that claim to support the ``/auth`` route and don't have data
            center specific credentials.
            You can also use a URL mapping as for the normal FDSN client
            instead of the URL.
        """
        HTTPClient.__init__(self, debug=debug, timeout=timeout)
        self.include_providers = include_providers
        self.exclude_providers = exclude_providers

        # Parse credentials.
        self.credentials = {}
        for key, value in (credentials or {}).items():
            if key == "EIDA_TOKEN":
                self.credentials[key] = value
            # Map, if necessary.
            if key in URL_MAPPINGS:
                key = URL_MAPPINGS[key]
            # Make sure urlparse works correctly.
            if not key.startswith("http"):
                key = "http://" + key
            # Only use the location.
            self.credentials[urlparse(key).netloc] = value

    @property
    def include_providers(self):
        return self.__include_providers

    @include_providers.setter
    def include_providers(self, value):
        self.__include_providers = self._expand_providers(value)

    @property
    def exclude_providers(self):
        return self.__exclude_providers

    @exclude_providers.setter
    def exclude_providers(self, value):
        self.__exclude_providers = self._expand_providers(value)

    def _expand_providers(self, providers):
        if providers is None:
            providers = []
        elif isinstance(providers, str):
            providers = [providers]
        return [_strip_protocol(URL_MAPPINGS[_i])
                if _i in URL_MAPPINGS
                else _strip_protocol(_i) for _i in providers]

    def _filter_requests(self, split):
        """
        Filter requests based on including and excluding providers.

        :type split: dict
        :param split: A dictionary containing the desired routing.
        """
        key_map = {_strip_protocol(url): url for url in split.keys()}

        # Apply both filters.
        f_keys = set(key_map.keys())
        if self.include_providers:
            f_keys = f_keys.intersection(set(self.include_providers))
        f_keys = f_keys.difference(set(self.exclude_providers))

        return {key_map[k]: split[key_map[k]] for k in f_keys}

    def _download_waveforms(self, split, **kwargs):
        return self._download_parallel(split, data_type="waveform", **kwargs)

    def _download_stations(self, split, **kwargs):
        return self._download_parallel(split, data_type="station", **kwargs)

    def _download_parallel(self, split, data_type, **kwargs):
        # Apply the provider filter.
        split = self._filter_requests(split)

        if not split:
            raise FDSNNoDataException(
                "Nothing remains to download after the provider "
                "inclusion/exclusion filters have been applied.")

        if data_type not in ["waveform", "station"]:  # pragma: no cover
            raise ValueError("Invalid data type.")

        # One thread per data center.
        dl_requests = []
        for k, v in split.items():
            dl_requests.append({
                "debug": self._debug,
                "timeout": self._timeout,
                "endpoint": k,
                "bulk_str": v,
                "data_type": data_type,
                "kwargs": kwargs,
                "credentials": self.credentials})
        pool = ThreadPool(processes=len(dl_requests))
        results = pool.map(_try_download_bulk, dl_requests)

        # Merge all results into a single object.
        if data_type == "waveform":
            collection = obspy.Stream()
        elif data_type == "station":
            collection = obspy.Inventory(
                networks=[],
                source="ObsPy FDSN Routing %s" % obspy.__version__)
        else:  # pragma: no cover
            raise ValueError

        for _i in results:
            if not _i:
                continue
            collection += _i

        # Explitly close the thread pool as somehow this does not work
        # automatically under linux. See #2342.
        pool.close()

        return collection

    def _handle_requests_http_error(self, r):
        """
        This assumes the same error code semantics as the base fdsnws web
        services.

        Please overwrite this method in a child class if necessary.
        """
        reason = r.reason.encode()
        if hasattr(r, "content"):
            reason += b" -- " + r.content
        with io.BytesIO(reason) as buf:
            raise_on_error(r.status_code, buf)

    @_assert_filename_not_in_kwargs
    @_assert_attach_response_not_in_kwargs
    def get_waveforms(self, starttime, endtime, **kwargs):
        """
        Get waveforms from multiple data centers.

        Arguments are the same as in
        :meth:`obspy.clients.fdsn.client.Client.get_waveforms()`.
        Any additional ``**kwargs`` are passed on to each individual service's
        dataselect service if the service supports them (otherwise they are
        silently ignored for that particular fdsnws endpoint).

        The ``filename`` and ``attach_response`` parameters of the single
        provider FDSN client are not supported.

        This can route on a number of different parameters, depending on the
        service, please see the web site of each individual routing service
        for details.
        """
        # This just calls the bulk downloader to only implement the logic once.
        # Just pass these to the bulk request.
        bulk = []
        for _i in ["network", "station", "location", "channel"]:
            if _i in kwargs:
                bulk.append(kwargs[_i])
                del kwargs[_i]
            else:
                bulk.append("*")
        bulk.extend([starttime, endtime])
        return self.get_waveforms_bulk([bulk], **kwargs)

    def get_service_version(self):
        """
        Return a semantic version number of the remote service as a string.
        """
        r = self._download(self._url + "/version")
        return r.content.decode() if \
            hasattr(r.content, "decode") else r.content

    @_assert_filename_not_in_kwargs
    def get_stations(self, **kwargs):
        """
        Get stations from multiple data centers.

        It will pass on most parameters to the underlying routed service.
        They will also be passed on to the individual FDSNWS implementations
        if a service supports them.

        The ``filename`` parameter of the single provider FDSN client is not
        supported.

        This can route on a number of different parameters, please see the
        web sites of the `EarthScope (former IRIS) Federator
        <https://service.earthscope.org/irisws/fedcatalog/1/>`_ and of the
        `EIDAWS Routing Service
        <http://www.orfeus-eu.org/data/eida/webservices/routing/>`_ for
        details.
        """
        # Just pass these to the bulk request.
        bulk = [kwargs.pop(key, '*') for key in (
                "network", "station", "location", "channel", "starttime",
                "endtime")]
        return self.get_stations_bulk([bulk], **kwargs)


if __name__ == '__main__':  # pragma: no cover
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
"""
obspy.io.ah - AH (Ad Hoc) read support for ObsPy
================================================
This module provides read support for the AH (Ad Hoc) waveform format
introduced by the Lamont-Doherty Geological Observatory.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)

"""

if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
"""
obspy.clients.fdsn.routing - Routing services for FDSN web services
===================================================================

:copyright:
    The ObsPy Development Team (devs@obspy.org)
    Celso G Reyes, 2017
    EarthScope/IRIS-DMC
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""

# Extremely ugly way to avoid a race condition the first time strptime is
# imported which is not thread safe...
#
# See https://bugs.python.org/issue7980
import time
time.strptime("2000/11/30", "%Y/%m/%d")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
A class parsing WADL files describing FDSN web services.

There are couple of datacenter specific fixes in here. They are marked by XXX
and should be removed once the datacenters are fully standard compliant.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import io
import warnings
from collections import defaultdict

from lxml import etree

from obspy import UTCDateTime
from .header import (DEFAULT_PARAMETERS, DEFAULT_TYPES,
                     WADL_PARAMETERS_NOT_TO_BE_PARSED)


class WADLParser(object):
    def __init__(self, wadl_string):
        doc = etree.parse(io.BytesIO(wadl_string)).getroot()
        self.nsmap = doc.nsmap
        self._ns = self.nsmap.get(None, None)
        self.parameters = {}
        self._has_eida_auth = False

        # Get the url.
        url = self._xpath(doc, "/application/resources")[0].get("base")
        if "dataselect" in url:
            self._wadl_type = "dataselect"
            self._has_eida_auth = self._check_for_eida_auth(doc)
        elif "station" in url:
            self._wadl_type = "station"
        elif "event" in url:
            self._wadl_type = "event"
        else:
            raise NotImplementedError
        self._default_parameters = DEFAULT_PARAMETERS[self._wadl_type]

        # Map short names to long names.
        self._short_to_long_mapping = {}
        for item in self._default_parameters:
            if len(item) == 1:
                continue
            self._short_to_long_mapping[item[1]] = item[0]

        # Retrieve all the parameters.
        parameters = self._xpath(doc, "//method[@name='GET']/request/param")

        # The following is an attempt to parse WADL files in a very general way
        # that is hopefully able to deal with faulty WADLs as maintaining a
        # list of special cases for different WADLs is not a good solution.
        all_parameters = defaultdict(list)

        # Group the parameters by the 'id' attribute of the grandparents tag.
        # The 'name' tag will always be 'GET' due to the construction of the
        # xpath expression.
        for param in parameters:
            gparent = param.getparent().getparent()
            id_attr = gparent.get("id") or ""
            all_parameters[id_attr.lower()].append(param)

        # If query is a key, choose it.
        if "query" in all_parameters:
            parameters = all_parameters["query"]
        # Otherwise discard any keys that have "auth" in them but choose others
        # that have query in them. If all of that fails but an empty "id"
        # attribute is available, choose that.
        else:
            for key in all_parameters.keys():
                if "query" in key and "auth" not in key:
                    parameters = all_parameters[key]
                    break
            else:
                if "" in all_parameters:
                    parameters = all_parameters[""]
                else:
                    msg = "Could not parse the WADL at '%s'. Invalid WADL?" \
                        % url
                    raise ValueError(msg)

        if not parameters:
            msg = "Could not find any parameters"
            raise ValueError(msg)

        for param in parameters:
            self.add_parameter(param)

        # Raise a warning if some default parameters are not specified.
        missing_params = []
        for param in self._default_parameters:
            if param not in self.parameters.keys():
                missing_params.append(param)
        if missing_params:
            msg = ("The '%s' service at '%s' cannot deal with the following "
                   "required parameters: %s\nThey will not be available "
                   "for any requests. Any attempt to use them will result "
                   "in an error.") % (self._wadl_type, url,
                                      ", ".join(missing_params))
            warnings.warn(msg)

    def add_parameter(self, param_doc):
        name = param_doc.get("name")

        # Map the short to the long names.
        if name in self._short_to_long_mapping:
            name = self._short_to_long_mapping[name]

        # Skip the parameter if should be ignored.
        if name in WADL_PARAMETERS_NOT_TO_BE_PARSED:
            return

        # XXX: Special handling for the USGS event WADL. minlongitude is preset
        # twice... Remove once they fix it.
        if name == "minlongitude" and "minlongitude" in self.parameters:
            name = "maxlongitude"

        style = param_doc.get("style")
        if style != "query":
            msg = "Unknown parameter style '%s' in WADL" % style
            warnings.warn(msg)

        required = self._convert_boolean(param_doc.get("required"))
        if required is None:
            required = False

        param_type = param_doc.get("type")
        if param_type is None:
            # If not given, choose one from the DEFAULT_TYPES dictionary.
            # Otherwise assign a string.
            if name in DEFAULT_TYPES:
                param_type = DEFAULT_TYPES[name]
            else:
                param_type = str
        else:
            p = param_type.lower()
            if "date" in p:
                param_type = UTCDateTime
            elif "string" in p:
                param_type = str
            elif "double" in p or "float" in p:
                param_type = float
            elif "long" in p or "int" in p:
                param_type = int
            elif "bool" in p:
                param_type = bool
            else:
                msg = "Unknown parameter type '%s' in WADL." % param_type
                raise ValueError(msg)

        default_value = param_doc.get("default")
        if default_value is not None:
            if param_type == bool:
                default_value = self._convert_boolean(default_value)
            else:
                default_value = param_type(default_value)

        # Parse any possible options.
        options = []
        for option in self._xpath(param_doc, "option"):
            options.append(param_type(option.get("value")))

        doc = ""
        doc_title = ""
        for doc_elem in self._xpath(param_doc, "doc"):
            title = doc_elem.get("title")
            body = doc_elem.text
            doc_title = title
            if body:
                doc = body
            break

        self.parameters[name] = {
            "required": required,
            "type": param_type,
            "options": options,
            "doc_title": doc_title and doc_title.strip(),
            "doc": doc and doc.strip(),
            "default_value": default_value}

    @staticmethod
    def _convert_boolean(boolean_string):
        """
        Helper function for boolean value conversion.

        >>> WADLParser._convert_boolean("true")
        True
        >>> WADLParser._convert_boolean("True")
        True

        >>> WADLParser._convert_boolean("false")
        False
        >>> WADLParser._convert_boolean("False")
        False

        >>> WADLParser._convert_boolean("something")
        >>> WADLParser._convert_boolean(1)
        """
        try:
            if boolean_string.lower() == "false":
                return False
            elif boolean_string.lower() == "true":
                return True
            else:
                return None
        except Exception:
            return None

    def _xpath(self, doc, expr):
        """
        Simple helper method for using xpaths with the default namespace.
        """
        nsmap = self.nsmap.copy()
        if self._ns is not None:
            default_abbreviation = "default"
            # being paranoid, can happen that another ns goes by that name
            while default_abbreviation in nsmap:
                default_abbreviation = default_abbreviation + "x"
            parts = []
            # insert prefixes for default namespace
            for x in expr.split("/"):
                if x != "" and ":" not in x:
                    x = "%s:%s" % (default_abbreviation, x)
                parts.append(x)
            expr = "/".join(parts)
            # adapt nsmap accordingly
            nsmap.pop(None, None)
            nsmap[default_abbreviation] = self._ns
        return doc.xpath(expr, namespaces=nsmap)

    def _check_for_eida_auth(self, doc):
        auth_endpoint = self._xpath(doc, '//resources/resource[@path="auth"]')
        return auth_endpoint and True or False


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
The tf_misfit test suite.
"""
import numpy as np
import pytest
from scipy.signal import hilbert
import matplotlib.pyplot as plt

from obspy.signal.tf_misfit import (eg, em, feg, fem, fpg, fpm, pg, pm, teg,
                                    tem, tfeg, tfem, tfpg, tfpm, tpg, tpm)
from obspy.signal.tf_misfit import plot_tfr, plot_tf_misfits, plot_tf_gofs


class TestTf:
    """
    Test cases for tf functions.
    """
    @pytest.fixture(scope='class')
    def state(self):
        """Return a dict with state needed to run tests."""
        tmax = 3.
        npts = 60
        dt = tmax / (npts - 1)

        fmin = 1.
        fmax = 3.
        nf = 3

        # Constants for s1
        a1 = 4.
        t1 = .1
        f1 = 2.
        phi1 = 0.

        # Constants for s1t and s1a
        ps = 0.1
        a1a = a1 * 1.1

        t = np.linspace(0., tmax, npts)
        f = np.logspace(np.log10(fmin), np.log10(fmax), nf)

        def h(t):
            return (np.sign(t) + 1) / 2

        def s1(t):
            return a1 * (t - t1) * np.exp(-2 * (t - t1)) * \
                np.cos(2. * np.pi * f1 * (t - t1) + phi1 * np.pi) * h(t - t1)

        # generate analytical signal (hilbert transform) and add phase shift
        s1h = hilbert(s1(t))
        s1p = np.real(
            np.abs(s1h) * np.exp(np.angle(s1h) * 1j + ps * np.pi * 1j))

        # signal with amplitude error
        def s1a(t):
            return a1a * (t - t1) * np.exp(-2 * (t - t1)) * \
                np.cos(2. * np.pi * f1 * (t - t1) + phi1 * np.pi) * h(t - t1)

        out = dict(
            s1=s1,
            s1p=s1p,
            s1a=s1a,
            t=t,
            f=f,
            dt=dt,
            fmin=fmin,
            fmax=fmax,
            nf=nf,
            npts=npts,
            w0=6
        )
        return out

    def test_phase_misfit(self, state, testdata):
        """
        Tests all tf misfits with a signal that has phase misfit
        """
        s1 = state['s1']
        s1p = state['s1p']
        t = state['t']
        dt = state['dt']

        fmin = state['fmin']
        fmax = state['fmax']
        nf = state['nf']

        tfem_11p_ref = np.loadtxt(testdata['TFEM_11p.dat'])
        tfpm_11p_ref = np.loadtxt(testdata['TFPM_11p.dat'])
        tem_11p_ref = np.loadtxt(testdata['TEM_11p.dat'])
        fem_11p_ref = np.loadtxt(testdata['FEM_11p.dat'])
        fpm_11p_ref = np.loadtxt(testdata['FPM_11p.dat'])
        tpm_11p_ref = np.loadtxt(testdata['TPM_11p.dat'])
        em_11p_ref = np.loadtxt(testdata['EM_11p.dat'])
        pm_11p_ref = np.loadtxt(testdata['PM_11p.dat'])

        tfem_11p = tfem(s1p, s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)
        tfpm_11p = tfpm(s1p, s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)
        tem_11p = tem(s1p, s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)
        fem_11p = fem(s1p, s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)
        fpm_11p = fpm(s1p, s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)
        tpm_11p = tpm(s1p, s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)
        em_11p = em(s1p, s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)
        pm_11p = pm(s1p, s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)

        tol = 1e-5
        atol_min = 1e-15

        assert np.allclose(tfem_11p, tfem_11p_ref, rtol=tol,
               atol=np.abs(tfem_11p_ref).max() * tol + atol_min)
        assert np.allclose(tfpm_11p, tfpm_11p_ref, rtol=tol,
               atol=np.abs(tfpm_11p_ref).max() * tol + atol_min)
        assert np.allclose(tem_11p, tem_11p_ref, rtol=tol,
               atol=np.abs(tem_11p_ref).max() * tol + atol_min)
        assert np.allclose(fem_11p, fem_11p_ref, rtol=tol,
               atol=np.abs(fem_11p_ref).max() * tol + atol_min)
        assert np.allclose(fpm_11p, fpm_11p_ref, rtol=tol,
               atol=np.abs(fpm_11p_ref).max() * tol + atol_min)
        assert np.allclose(tpm_11p, tpm_11p_ref, rtol=tol,
               atol=np.abs(tpm_11p_ref).max() * tol + atol_min)
        assert np.allclose(em_11p, em_11p_ref, rtol=tol,
               atol=np.abs(em_11p_ref).max() * tol + atol_min)
        assert np.allclose(pm_11p, pm_11p_ref, rtol=tol,
               atol=np.abs(pm_11p_ref).max() * tol + atol_min)

        # keeping the save commands in case the files need to be updated
        # np.savetxt(self.path + os.sep + 'TFEM_11p.dat', TFEM_11p,
        #            fmt='%1.5e')
        # np.savetxt(self.path + os.sep + 'TFPM_11p.dat', TFPM_11p,
        #            fmt='%1.5e')
        # np.savetxt(self.path + os.sep + 'TEM_11p.dat', TEM_11p, fmt='%1.5e')
        # np.savetxt(self.path + os.sep + 'FEM_11p.dat', FEM_11p, fmt='%1.5e')
        # np.savetxt(self.path + os.sep + 'FPM_11p.dat', FPM_11p, fmt='%1.5e')
        # np.savetxt(self.path + os.sep + 'TPM_11p.dat', TPM_11p, fmt='%1.5e')
        # np.savetxt(self.path + os.sep + 'EM_11p.dat', (EM_11p,), fmt='%1.5e')
        # np.savetxt(self.path + os.sep + 'PM_11p.dat', (PM_11p,), fmt='%1.5e')

    def test_envelope_misfit(self, state, testdata):
        """
        Tests all tf misfits with a signal that has envelope misfit
        """
        s1 = state['s1']
        s1a = state['s1a']
        t = state['t']
        dt = state['dt']

        fmin = state['fmin']
        fmax = state['fmax']
        nf = state['nf']

        tfem_11a_ref = np.loadtxt(testdata['TFEM_11a.dat'])
        tfpm_11a_ref = np.loadtxt(testdata['TFPM_11a.dat'])
        tem_11a_ref = np.loadtxt(testdata['TEM_11a.dat'])
        fem_11a_ref = np.loadtxt(testdata['FEM_11a.dat'])
        fpm_11a_ref = np.loadtxt(testdata['FPM_11a.dat'])
        tpm_11a_ref = np.loadtxt(testdata['TPM_11a.dat'])
        em_11a_ref = np.loadtxt(testdata['EM_11a.dat'])
        pm_11a_ref = np.loadtxt(testdata['PM_11a.dat'])

        tfem_11a = tfem(s1a(t), s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)
        tfpm_11a = tfpm(s1a(t), s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)
        tem_11a = tem(s1a(t), s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)
        tpm_11a = tpm(s1a(t), s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)
        fem_11a = fem(s1a(t), s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)
        fpm_11a = fpm(s1a(t), s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)
        em_11a = em(s1a(t), s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)
        pm_11a = pm(s1a(t), s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)

        tol = 1e-5
        atol_min = 1e-15

        assert np.allclose(tfem_11a, tfem_11a_ref, rtol=tol,
               atol=np.abs(tfem_11a_ref).max() * tol + atol_min)
        assert np.allclose(tfpm_11a, tfpm_11a_ref, rtol=tol,
               atol=np.abs(tfpm_11a_ref).max() * tol + atol_min)
        assert np.allclose(tem_11a, tem_11a_ref, rtol=tol,
               atol=np.abs(tem_11a_ref).max() * tol + atol_min)
        assert np.allclose(fem_11a, fem_11a_ref, rtol=tol,
               atol=np.abs(fem_11a_ref).max() * tol + atol_min)
        assert np.allclose(fpm_11a, fpm_11a_ref, rtol=tol,
               atol=np.abs(fpm_11a_ref).max() * tol + atol_min)
        assert np.allclose(tpm_11a, tpm_11a_ref, rtol=tol,
               atol=np.abs(tpm_11a_ref).max() * tol + atol_min)
        assert np.allclose(em_11a, em_11a_ref, rtol=tol,
               atol=np.abs(em_11a_ref).max() * tol + atol_min)
        assert np.allclose(pm_11a, pm_11a_ref, rtol=tol,
               atol=np.abs(pm_11a_ref).max() * tol + atol_min)

        # keeping the save commands in case the files need to be updated
        # np.savetxt(self.path + os.sep + 'TFEM_11a.dat', TFEM_11a,
        #            fmt='%1.5e')
        # np.savetxt(self.path + os.sep + 'TFPM_11a.dat', TFPM_11a,
        #            fmt='%1.5e')
        # np.savetxt(self.path + os.sep + 'TEM_11a.dat', TEM_11a, fmt='%1.5e')
        # np.savetxt(self.path + os.sep + 'FEM_11a.dat', FEM_11a, fmt='%1.5e')
        # np.savetxt(self.path + os.sep + 'FPM_11a.dat', FPM_11a, fmt='%1.5e')
        # np.savetxt(self.path + os.sep + 'TPM_11a.dat', TPM_11a, fmt='%1.5e')
        # np.savetxt(self.path + os.sep + 'EM_11a.dat', (EM_11a,), fmt='%1.5e')
        # np.savetxt(self.path + os.sep + 'PM_11a.dat', (PM_11a,), fmt='%1.5e')

    def test_envelope_gof(self, state):
        """
        Tests all tf gofs
        """
        s1 = state['s1']
        t = state['t']
        dt = state['dt']

        fmin = state['fmin']
        fmax = state['fmax']
        nf = state['nf']
        npts = state['npts']

        tol = 1e-5

        _tfeg = tfeg(s1(t), s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)
        _tfpg = tfpg(s1(t), s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)
        _teg = teg(s1(t), s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)
        _tpg = tpg(s1(t), s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)
        _feg = feg(s1(t), s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)
        _fpg = fpg(s1(t), s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)
        _eg = eg(s1(t), s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)
        _pg = pg(s1(t), s1(t), dt=dt, fmin=fmin, fmax=fmax, nf=nf)

        assert np.allclose(_tfeg, np.ones((nf, npts)) * 10., rtol=tol)
        assert np.allclose(_tfpg, np.ones((nf, npts)) * 10., rtol=tol)
        assert np.allclose(_teg, np.ones(npts) * 10., rtol=tol)
        assert np.allclose(_tpg, np.ones(npts) * 10., rtol=tol)
        assert np.allclose(_feg, np.ones(nf) * 10., rtol=tol)
        assert np.allclose(_fpg, np.ones(nf) * 10., rtol=tol)
        assert np.allclose(_eg, 10., rtol=tol)
        assert np.allclose(_pg, 10., rtol=tol)


class TestTfPlot:
    """
    Test cases for tf plot functions.
    """
    @pytest.fixture(scope='class')
    def state(self):
        """return state for testing."""
        from obspy.core.util import AttribDict
        out = AttribDict()
        # general constants
        tmax = 6.
        out.dt = 0.01
        npts = int(tmax / out.dt + 1)
        t = np.linspace(0., tmax, npts)
        out.fmin = .5
        out.fmax = 10
        # constants for the signal
        a1 = 4.
        t1 = 2.
        f1 = 2.
        phi1 = 0.
        # amplitude and phase error
        phase_shift = 0.1
        amp_fac = 1.2

        # generate the signal
        h1 = (np.sign(t - t1) + 1) / 2
        out.st1 = a1 * (t - t1) * np.exp(-2 * (t - t1))
        out.st1 *= np.cos(2. * np.pi * f1 * (t - t1) + phi1 * np.pi) * h1

        # reference signal
        out.st2 = out.st1.copy()

        # generate analytical signal (hilbert transform) and add phase shift
        out.st1p = hilbert(out.st1)
        out.st1p = np.real(
            np.abs(out.st1p) *
            np.exp((np.angle(out.st1p) + phase_shift * np.pi) * 1j))

        # signal with amplitude error
        out.st1p /= amp_fac
        return out

    def test_plot_tfr(self, state, ignore_numpy_errors, image_path):
        n = 295
        t, dt = np.linspace(0., 20 * np.pi, n, retstep=True)
        sig = np.sin(t)
        plot_tfr(sig, dt=dt, show=False)
        plt.savefig(image_path)

    def test_plot_tf_misfits(self, image_path, ignore_numpy_errors, state):
        plot_tf_misfits(state.st1p, state.st2, dt=state.dt,
                        fmin=state.fmin, fmax=state.fmax, show=False)
        plt.savefig(image_path)

    def test_plot_tf_gofs(self, state, ignore_numpy_errors, image_path):
        plot_tf_gofs(state.st1p, state.st2, dt=state.dt, fmin=state.fmin,
                     fmax=state.fmax, show=False)
        plt.savefig(image_path)



# -*- coding: utf-8 -*-
"""
Ray path calculations.
"""
from .taup_pierce import TauPPierce


class TauPPath(TauPPierce):
    """
    Calculate the ray paths for each phase, using TauPPierce and TauPTime.
    """
    def calculate(self, degrees):
        """
        Call all the necessary calculations to obtain the ray paths.
        """
        self.depth_correct(self.source_depth, self.receiver_depth)
        self.recalc_phases()
        self.arrivals = []
        self.calculate_path(degrees)

    def calculate_path(self, degrees):
        """
        Calculates the ray paths for phases at the given distance by
        calling the calculate_path method of the SeismicPhase class. The
        results are then in self.arrivals.
        """
        self.degrees = degrees
        for phase in self.phases:
            self.arrivals += phase.calc_path(degrees, self.ray_param_tol)
